import argparse
import os
import subprocess
import pandas as pd
import glob
import shutil
import re
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_03_do_clustering")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument(
        "--min_ani",
        type=int,
        default=95,
        help="Minimum ANI value for clustering (default = 95)",
    )
    parser.add_argument(
        "--min_tcov",
        type=int,
        default=85,
        help="Minimum coverage of the target sequence (default = 85)",
    )
    parser.add_argument(
        "--min_qcov",
        type=int,
        default=0,
        help="Minimum coverage of the query sequence (default = 0)",
    )
    parser.add_argument(
        "--read-type",
        type=str,
        default='short',
        help="Sequencing data type (e.g. short vs long reads). Default = short",
    )
    parser.add_argument(
        "--unfiltered_protein_file",
        action="store_true",
        default=False,
        help="Create protein FASTA file from unfiltered virus sequence. Default = False. Warning = If argument provided, the script might run for a long period of time.",
    )
    parser.add_argument(
        '--threads', 
        type=int,
        default=1,
        help='Number of threads to use (default = 1)')

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()

def rename_sequence(sequence_name):
    sequence_name = sequence_name.replace('provirus_', '')
    sequence_name = sequence_name.rsplit('/', 1)[0]
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')
    return sequence_name

def rename_protein_sequence(sequence_name):
    original_sequence_name = sequence_name  # Store the original sequence name
    sequence_name = sequence_name.replace('provirus_', '')
    if ' ' in sequence_name:
        sequence_name, additional_info = sequence_name.split(' ', 1)
    else:
        additional_info = ''

    sequence_name = re.sub(r'\/\w+_(\w+)', r'_\1', sequence_name)
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')

    additional_parts = additional_info.split(' ')
    if len(additional_parts) > 1:
        additional_parts[0] = '_'.join(part.strip() for part in additional_parts[0].split('_'))

    modified_sequence_name = f"{sequence_name} {' '.join(additional_parts)}"

    return f"{modified_sequence_name} {original_sequence_name}"

def rename_protein_sequence_table(sequence_name):
    sequence_name = sequence_name.replace('provirus_', '')
    sequence_name = re.sub(r'\/\w+_(\w+)', r'_\1', sequence_name)
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')
    return sequence_name

def generate_output_tables(args, rename_sequence):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating the merge geNomad and CheckV output tables...\033[0m")
    # Search for files ending with '_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv'
    unfiltered_quality_summary_files = []
    unfiltered_quality_summary_file_pattern = os.path.join(args["input"], '02_CHECK_V', '*', '*_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    unfiltered_quality_summary_files.extend(glob.glob(unfiltered_quality_summary_file_pattern))

    # Initialize an empty DataFrame to store the concatenated data
    unfiltered_quality_summary_concatenated_df = pd.DataFrame()

    # Loop through each quality summary file and concatenate the data
    for unfiltered_quality_summary_file in unfiltered_quality_summary_files:
        unfiltered_quality_summary_df = pd.read_csv(unfiltered_quality_summary_file, sep='\t')
        unfiltered_quality_summary_concatenated_df = pd.concat([unfiltered_quality_summary_concatenated_df, unfiltered_quality_summary_df], ignore_index=True)

    # Save the quality_summary_concatenated_df DataFrame to a new file
    unfiltered_quality_summary_concatenated_path = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    unfiltered_quality_summary_concatenated_df['virus_id'] = unfiltered_quality_summary_concatenated_df['virus_id'].apply(rename_sequence)
    unfiltered_quality_summary_concatenated_df.to_csv(unfiltered_quality_summary_concatenated_path, sep='\t', index=False)

    # Search for files ending with '_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv'
    filtered_quality_summary_files = []
    filtered_quality_summary_file_pattern = os.path.join(args["input"], '02_CHECK_V', '*', '*_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    filtered_quality_summary_files.extend(glob.glob(filtered_quality_summary_file_pattern))

    # Initialize an empty DataFrame to store the concatenated data
    filtered_quality_summary_concatenated_df = pd.DataFrame()

    # Loop through each quality summary file and concatenate the data
    for filtered_quality_summary_file in filtered_quality_summary_files:
        filtered_quality_summary_df = pd.read_csv(filtered_quality_summary_file, sep='\t')
        filtered_quality_summary_concatenated_df = pd.concat([filtered_quality_summary_concatenated_df, filtered_quality_summary_df], ignore_index=True)

    # Save the quality_summary_concatenated_df DataFrame to a new file
    filtered_quality_summary_concatenated_path = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    filtered_quality_summary_concatenated_df.to_csv(filtered_quality_summary_concatenated_path, sep='\t', index=False)

    return unfiltered_quality_summary_concatenated_df, filtered_quality_summary_concatenated_df

def clustering(args, filtered_quality_summary_concatenated_df):
    print(f"\n\033[1m{step_counter.print_sub_step()}: Concatenating all FASTA sequence files...\033[0m")
    # Get a list of all input FASTA files within the specified directories
    filtered_sequence_files = []
    filtered_sequence_file_pattern = os.path.join(args["input"], '02_CHECK_V', '*', '*_Filtered_Relaxed_Virus_Provirus_Sequences.fna')
    filtered_sequence_files.extend(glob.glob(filtered_sequence_file_pattern, recursive=True))

    # Concatenate the input FASTA files
    output_file = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences.fna')
    with open(output_file, 'w') as outfile:
        for fasta_file in filtered_sequence_files:
            with open(fasta_file, 'r') as infile:
                outfile.write(infile.read())

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.abspath(__file__))
    # Construct the path to 03_anicalc.py and 03_aniclust.py
    anicalc_path = os.path.join(script_dir, '03_anicalc.py')
    aniclust_path = os.path.join(script_dir, '03_aniclust.py')

    print(f"\n\033[1m{step_counter.print_sub_step()}: Running makeblastdb on the concatenated FASTA file...\033[0m")
    blastdb_output = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_BLAST_DB')
    subprocess.run(['makeblastdb', '-in', output_file, '-dbtype', 'nucl', '-out', blastdb_output])

    print(f"\n\033[1m{step_counter.print_sub_step()}: Running blastn on the concatenated FASTA file...\033[0m")
    blastn_output = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_Clustering.tsv')
    subprocess.run(['blastn', '-query', output_file, '-db', blastdb_output, '-outfmt', '6 std qlen slen', '-max_target_seqs', '10000', '-out', blastn_output, '-num_threads', str(args["threads"])])

    print(f"\n\033[1m{step_counter.print_sub_step()}: Creating the clustering output table containing the Representative sequences...\033[0m")
    # Run anicalc.py on the blastn output
    ani_output = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_Clustering_ANI.tsv')
    subprocess.run(['python', anicalc_path, '-i', blastn_output, '-o', ani_output])

    # Run aniclust.py on the anicalc.py output
    cluster_output = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_Clustering_ANI_Clusters.tsv')
    subprocess.run(['python', aniclust_path, '--fna', output_file, '--ani', ani_output, '--out', cluster_output, '--min_ani', str(args["min_ani"]), '--min_tcov', str(args["min_tcov"]), '--min_qcov', str(args["min_qcov"])])

    # Add column names to the output file
    df = pd.read_csv(cluster_output, sep='\t', header=None)
    df.columns = ['Representative_Sequence', 'Sequences']
    contig_ids_to_keep = set(df['Representative_Sequence'].tolist())
    df.to_csv(cluster_output, sep='\t', index=False)

    representative_filtered_quality_summary_concatenated_df = pd.merge(df, filtered_quality_summary_concatenated_df, left_on='Representative_Sequence',right_on='virus_id', how='left')
    columns_to_exclude = ['virus_id']  # Specify the columns to exclude
    representative_filtered_quality_summary_concatenated_df = representative_filtered_quality_summary_concatenated_df.drop(columns=columns_to_exclude)

    # Save the merged DataFrame to a new file
    representative_filtered_quality_summary_concatenated_path = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Representative_Virus_Proviruses_Quality_Summary.tsv')
    representative_filtered_quality_summary_concatenated_df.to_csv(representative_filtered_quality_summary_concatenated_path, sep='\t', index=False)

    return output_file, contig_ids_to_keep, blastn_output, ani_output, cluster_output, representative_filtered_quality_summary_concatenated_df

def create_representative_fasta_file(args, output_file, contig_ids_to_keep):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the FASTA file containing the representative sequences from the clustering step...\033[0m")
    # Find the directory containing the viruses and proviruses fasta files for this sample
    output_file = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences.fna')

    # Read all lines from the concatenated file
    fasta_lines = []
    with open(output_file, 'r') as f:
        fasta_lines = f.readlines()

    # Filter the concatenated file
    representative_fasta_file = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Representative_Virus_Provirus_Sequences.fna')
    with open(representative_fasta_file, 'w') as f:
        write_line = False
        for line in fasta_lines:
            if line.startswith('>'):
                contig_id = line.strip()[1:]
                if contig_id in contig_ids_to_keep:
                    write_line = True
                    f.write(line)
                else:
                    write_line = False
            else:
                if write_line:
                    f.write(line)

    return representative_fasta_file

def clean_folder(args, blastn_output, ani_output, cluster_output):
    print(f"\n\033[1m{step_counter.print_main_step()}: Cleaning 03_CLUSTERING directory by removing intermediary files...\033[0m")
    # Create 'tmp' folder
    tmp_folder = os.path.join(args["input"], '03_CLUSTERING', 'tmp')

    # Delete the existing 'tmp' folder if it exists
    if os.path.exists(tmp_folder):
        shutil.rmtree(tmp_folder)

    # Create a new 'tmp' folder
    os.makedirs(tmp_folder, exist_ok=True)

    # Define the source directory and file pattern
    source_directory = os.path.join(args["input"], '03_CLUSTERING')
    file_pattern = 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_BLAST_DB.*'

    # Use glob to find all matching files
    matching_files = glob.glob(os.path.join(source_directory, file_pattern))

    # Move matching files to the temporary folder
    for file_path in matching_files:
        shutil.move(file_path, tmp_folder)

    # Move other files to the temporary folder
    shutil.move(blastn_output, tmp_folder)
    shutil.move(ani_output, tmp_folder)
    shutil.move(cluster_output, tmp_folder)

def generate_index(args, representative_fasta_file):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the reference index from the FASTA file containing the representative sequences for Module 04 (Read mapping)...\033[0m")
    # Find path for '04_READ_MAPPING'
    read_mapping_dir = os.path.join(args["input"], '04_READ_MAPPING')

    # If args.read_type is "short" or not provided, create a Bowtie2 index if it does not exist
    reference_path = os.path.join(read_mapping_dir, 'reference')

    if args["read_type"] == 'short' or not args["read_type"]:
        if not os.path.exists(reference_path + '.1.bt2'):
            print("\nRun bowtie2-build: use representative viral sequences to build bowtie database.\n")
            subprocess.run(['bowtie2-build', representative_fasta_file, reference_path])
        else:
            print(f"\nReference index 'reference' already exists in {read_mapping_dir}")
    else:
        # If args.read_type is "long," apply minimap2 to create an index (assuming minimap2 is in your PATH)
        if not os.path.exists(reference_path + '.mmi'):
            print("\nRun minimap2 to create an index from representative viral sequences.\n")
            subprocess.run(['minimap2', '-d', reference_path, representative_fasta_file])
        else:
            print(f"\nReference index 'reference' already exists in {read_mapping_dir}")


def generate_protein_files(args, metadata, output_file, representative_fasta_file):
    functional_annotation_directory = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION')
    os.makedirs(functional_annotation_directory, exist_ok=True)

    # Create a dictionary to store sequences by sample_name
    sequences_by_sample = {}

    # Iterate over rows in the metadata DataFrame
    for row_index, (_, row) in enumerate(metadata.iterrows(), start=1):
        sample_name = row['Sample']
        # Construct the path to the provirus protein file based on the sample name
        provirus_protein_file = os.path.join(args["input"], '01_GENOMAD', str(sample_name), f'{sample_name}_Proviruses_Genomad_Output', 'proviruses_summary', 'proviruses_virus_proteins.faa')
        checkv_results = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Viruses_CheckV_Output', 'quality_summary.tsv')
        virus_protein_files = glob.glob(os.path.join(args["input"], '01_GENOMAD', str(sample_name), f'{sample_name}_Viruses_Genomad_Output', f'*_summary', f'*_virus_proteins.faa'))
        checkv_df_proviruses = pd.read_csv(checkv_results, sep='\t')
        list_proviruses = checkv_df_proviruses.loc[checkv_df_proviruses["provirus"] == "Yes"]['contig_id'].to_list()

        # Check if the provirus protein file exists
        if os.path.exists(provirus_protein_file):

            # Load sequences from the provirus protein file
            provirus_sequences = SeqIO.to_dict(SeqIO.parse(provirus_protein_file, "fasta"))

            # Initialize a list to store sequences from virus_protein_files that are not similar to provirus_sequences
            non_similar_sequences = []

            # Iterate over virus protein files
            for virus_protein_file in virus_protein_files:
                # Load sequences from the virus protein file
                virus_sequences = SeqIO.to_dict(SeqIO.parse(virus_protein_file, "fasta"))

                # Find sequences from virus_sequences that are not similar to any sequences in provirus_sequences
                for seq_id, seq_record in virus_sequences.items():
                    contig_id = "_".join(seq_id.split("_")[0:-1])
                    if contig_id not in list_proviruses:
                        non_similar_sequences.append(seq_record)

            # Store the combined sequences from provirus_protein_file and non_similar_sequences
            combined_sequences = list(provirus_sequences.values()) + non_similar_sequences
            sequences_by_sample[sample_name] = combined_sequences
        else:
            # If provirus protein file is missing, use virus protein files as is
            combined_sequences = []
            for virus_protein_file in virus_protein_files:
                combined_sequences.extend(SeqIO.parse(virus_protein_file, "fasta"))
            sequences_by_sample[sample_name] = combined_sequences

    if args["unfiltered_protein_file"]:
        print(f"\n\033[1m{step_counter.print_sub_step()}: Generating FASTA file containing protein sequences from unfiltered viruses...\033[0m")
        # Concatenate all sequences for each sample into a single file
        protein_file = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Unfiltered_Virus_Provirus_Protein_Sequences.faa")
        with open(protein_file, "w") as output_handle:
            for sample_name, sequences in sequences_by_sample.items():
                for seq_record in sequences:
                    # Rename the sequence header
                    seq_record.id = rename_protein_sequence(seq_record.id)
                    SeqIO.write(seq_record, output_handle, "fasta")

        return protein_file


    print(f"\n\033[1m{step_counter.print_sub_step()}: Generating FASTA file containing protein sequences from filtered (relaxed mode) viruses...\033[0m")
    # Load the headers from the representative fasta file into a set for faster lookup
    filtered_headers = set(SeqIO.to_dict(SeqIO.parse(output_file, "fasta")).keys())
    filtered_protein_file = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Filtered_Relaxed_Virus_Provirus_Protein_Sequences.faa")

    # Concatenate all sequences for each sample into a single file
    with open(filtered_protein_file, "w") as output_handle:
        for sample_name, sequences in sequences_by_sample.items():
            for seq_record in sequences:
                # Rename the sequence header
                seq_record.id = rename_protein_sequence(seq_record.id)
                # Check if the header is in the representative_headers set
                if any(substring in seq_record.id for substring in filtered_headers):
                    SeqIO.write(seq_record, output_handle, "fasta")

    
    # Load the headers from the representative fasta file into a set for faster lookup
    representative_headers = set(SeqIO.to_dict(SeqIO.parse(representative_fasta_file, "fasta")).keys())
    representative_protein_file = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Filtered_Relaxed_Representative_Virus_Provirus_Protein_Sequences.faa")

    print(f"\n\033[1m{step_counter.print_sub_step()}: Generating FASTA file containing protein sequences from filtered (relaxed mode) representative vOTUs...\033[0m")
    with open(representative_protein_file, "w") as output_handle:
        for sample_name, sequences in sequences_by_sample.items():
            for seq_record in sequences:
                # Rename the sequence header
                seq_record.id = rename_protein_sequence(seq_record.id)
                # Check if the header is in the representative_headers set
                if any(substring in seq_record.id for substring in representative_headers):
                    SeqIO.write(seq_record, output_handle, "fasta")


# Define a custom sorting key function
def custom_sort_key(gene):
    # Split the viral_gene_id into parts using "_"
    parts = gene.split('_')
    Contig_name = '_'.join(parts[:-1])  # Join all parts except the last one to get Contig_name
    gene_number = int(parts[-1]) if len(parts) > 1 else 0  # Extract the last part as Gene_number

    return Contig_name, gene_number

def generate_genomad_annotation_tables(args, protein_file, filtered_quality_summary_concatenated_df, representative_filtered_quality_summary_concatenated_df):
    # Get a list of all input gene annotation TSV file files within the specified directories
    gene_annotation_files = []
    gene_annotation_file_pattern = os.path.join(args["input"], '01_GENOMAD', '*', '*_Genomad_Output', '*_summary', '*_virus_genes.tsv')
    gene_annotation_files.extend(glob.glob(gene_annotation_file_pattern, recursive=True))
    
    # Concatenate all DataFrames into a single DataFrame
    dfs = []
    for file in gene_annotation_files:
        df = pd.read_csv(file, sep='\t', usecols=['gene', 'start', 'end', 'length', 'strand', 'gc_content', 'genetic_code', 'rbs_motif', 'marker', 'evalue', 
                                                'bitscore', 'virus_hallmark', 'annotation_amr', 'annotation_accessions',
                                                'annotation_description'])
        dfs.append(df)

    concatenated_gene_annotation_files = pd.concat(dfs, ignore_index=True)
    concatenated_gene_annotation_files['gene'] = concatenated_gene_annotation_files['gene'].apply(rename_protein_sequence_table)

    # Sort rows by Contig_name and Gene_number using the custom sorting key
    concatenated_gene_annotation_files[['Contig_name', 'Gene_number']] = concatenated_gene_annotation_files['gene'].apply(custom_sort_key).apply(pd.Series)
    concatenated_gene_annotation_files.insert(1, 'Contig_name', concatenated_gene_annotation_files.pop('Contig_name'))
    concatenated_gene_annotation_files.insert(2, 'Gene_number', concatenated_gene_annotation_files.pop('Gene_number'))

    if args["unfiltered_protein_file"]:
        print(f"\n\033[1m{step_counter.print_sub_step()}: Generating geNomad function annotation table containing unfiltered viral contigs...\033[0m")
        concatenated_gene_annotation_files_path = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv")
        protein_file_headers = set(SeqIO.to_dict(SeqIO.parse(protein_file, "fasta")).keys())
        concatenated_gene_annotation_files = concatenated_gene_annotation_files[concatenated_gene_annotation_files['gene'].isin(protein_file_headers)]
        concatenated_gene_annotation_files.to_csv(concatenated_gene_annotation_files_path, sep ='\t')


    print(f"\n\033[1m{step_counter.print_sub_step()}: Generating geNomad function annotation table containing filtered (relaxed mode) viral contigs...\033[0m")
    merged_concatenated_gene_annotation_files = pd.merge(filtered_quality_summary_concatenated_df[['virus_id']], concatenated_gene_annotation_files, left_on='virus_id',right_on='Contig_name', how='left')

    merged_concatenated_gene_annotation_files = merged_concatenated_gene_annotation_files.sort_values(by=['Contig_name', 'Gene_number'], ascending=[True, True])
    columns_to_fill_unknown = ['annotation_amr', 'annotation_accessions', 'annotation_description']
    merged_concatenated_gene_annotation_files[columns_to_fill_unknown] = merged_concatenated_gene_annotation_files[columns_to_fill_unknown].fillna('Unknown')
    merged_concatenated_gene_annotation_files.fillna('NA', inplace=True)
    merged_concatenated_gene_annotation_files.drop(columns=['virus_id'], inplace=True)
    merged_concatenated_gene_annotation_files.rename(columns={'gene': 'Viral_gene_ID'}, inplace=True)
    merged_concatenated_gene_annotation_files['virus_hallmark'] = merged_concatenated_gene_annotation_files['virus_hallmark'].replace({0: 'NA', 1: 'Virus_hallmark'})
    desired_column_order = ['Viral_gene_ID', 'Contig_name', 'Gene_number', 'start', 'end', 'length', 'strand', 'annotation_description', 'annotation_accessions', 'bitscore', 'evalue', 'gc_content', 'genetic_code', 'rbs_motif',
        'marker', 'virus_hallmark', 'annotation_amr']
    merged_concatenated_gene_annotation_files = merged_concatenated_gene_annotation_files[desired_column_order]
    new_column_names = {'gc_content': 'GENOMAD_gc_content', 'genetic_code': 'GENOMAD_genetic_code', 'rbs_motif': 'GENOMAD_rbs_motif', 'marker': 'GENOMAD_marker', 'evalue': 'GENOMAD_evalue',
        'bitscore': 'GENOMAD_Score',  'virus_hallmark': 'GENOMAD_virus_hallmark', 'annotation_amr': 'GENOMAD_Annotation_amr', 'annotation_accessions': 'GENOMAD_Annotation_accessions',
        'annotation_description': 'GENOMAD_Annotation'}
    merged_concatenated_gene_annotation_files.rename(columns=new_column_names, inplace=True)
    merged_concatenated_gene_annotation_files_path = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv")
    merged_concatenated_gene_annotation_files.to_csv(merged_concatenated_gene_annotation_files_path, sep ='\t')

    print(f"\n\033[1m{step_counter.print_sub_step()}: Generating geNomad function annotation table containing filtered (relaxed mode) representative vOTUs...\033[0m")
    representative_merged_concatenated_gene_annotation_files = pd.merge(representative_filtered_quality_summary_concatenated_df[['Representative_Sequence']], merged_concatenated_gene_annotation_files, left_on='Representative_Sequence',right_on='Contig_name', how='left')
    representative_merged_concatenated_gene_annotation_files.drop(columns=['Representative_Sequence'], inplace=True)
    representative_merged_concatenated_gene_annotation_files_path = os.path.join(args["input"], '06_FUNCTIONAL_ANNOTATION', "MVP_06_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Representative_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv")
    representative_merged_concatenated_gene_annotation_files.to_csv(representative_merged_concatenated_gene_annotation_files_path, sep ='\t')


def generate_summary_report(args, unfiltered_quality_summary_concatenated_df, filtered_quality_summary_concatenated_df, representative_filtered_quality_summary_concatenated_df, mvp_formatted_start_time, mvp_formatted_end_time, mvp_duration):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the summary report for Module 03 (Clustering)...\033[0m")
    # Calculate summary statistics before both filtration and clustering 
    num_unfiltered_viruses = len(unfiltered_quality_summary_concatenated_df)
    num_unfiltered_proviruses = (unfiltered_quality_summary_concatenated_df['provirus'].str.contains('Yes', case=False)).sum()
    mean_unfiltered_virus_length = unfiltered_quality_summary_concatenated_df['virus_length'].mean()
    max_unfiltered_virus_length = unfiltered_quality_summary_concatenated_df['virus_length'].max()
    min_unfiltered_virus_length = unfiltered_quality_summary_concatenated_df['virus_length'].min()
    unfiltered_virus_quality_counts = unfiltered_quality_summary_concatenated_df['checkv_quality'].value_counts()

    # Count occurrences of each realm in the second level of the taxonomy column
    unfiltered_virus_taxa_levels = unfiltered_quality_summary_concatenated_df['taxonomy'].str.split(';')
    unfiltered_virus_second_level_taxa = unfiltered_virus_taxa_levels.apply(lambda x: x[1] if len(x) > 1 else 'Unclassified/others')
    unfiltered_virus_taxa_counts = unfiltered_virus_second_level_taxa.value_counts()
    # Calculate the percentage of each realm
    unfiltered_virus_taxa_percentages = (unfiltered_virus_taxa_counts / num_unfiltered_viruses) * 100
    # Prepare the summary report content
    unfiltered_virus_taxa_summary = "\n".join([f"{realm}: {count} ({percentage:.1f}%)" for realm, count, percentage in zip(unfiltered_virus_taxa_counts.index, unfiltered_virus_taxa_counts.values, unfiltered_virus_taxa_percentages.values)])

    # Prepare the summary report content for top classes
    unfiltered_virus_taxa_levels = unfiltered_quality_summary_concatenated_df['taxonomy'].str.split(';').apply(lambda levels: levels + [''] * (5 - len(levels)))
    unfiltered_virus_fifth_level_classes = unfiltered_virus_taxa_levels.apply(lambda x: x[4])
    unfiltered_virus_filtered_classes = unfiltered_virus_fifth_level_classes[(unfiltered_virus_fifth_level_classes != '') & (unfiltered_virus_fifth_level_classes != 'Unclassified')]
    unfiltered_virus_class_counts = unfiltered_virus_filtered_classes.value_counts()
    unfiltered_virus_top_10_classes = unfiltered_virus_class_counts.nlargest(10)
    unfiltered_virus_top_classes_summary = "\n".join([f"{cls}: {count}" for cls, count in unfiltered_virus_top_10_classes.items()])

    # Calculate summary statistics after filtration and before clustering
    num_filtered_virus = len(filtered_quality_summary_concatenated_df)
    num_filtered_provirus = (filtered_quality_summary_concatenated_df['provirus'].str.contains('Yes', case=False)).sum()
    mean_filtered_virus_length = filtered_quality_summary_concatenated_df['virus_length'].mean()
    max_filtered_virus_length = filtered_quality_summary_concatenated_df['virus_length'].max()
    min_filtered_virus_length = filtered_quality_summary_concatenated_df['virus_length'].min()
    filtered_virus_quality_counts = filtered_quality_summary_concatenated_df['checkv_quality'].value_counts()

    # Count occurrences of each realm in the second level of the taxonomy column
    filtered_virus_taxa_levels = filtered_quality_summary_concatenated_df['taxonomy'].str.split(';')
    filtered_virus_second_level_taxa = filtered_virus_taxa_levels.apply(lambda x: x[1] if len(x) > 1 else 'Unclassified/others')
    filtered_virus_taxa_counts = filtered_virus_second_level_taxa.value_counts()
    # Calculate the percentage of each realm
    filtered_virus_taxa_percentages = (filtered_virus_taxa_counts / num_filtered_virus) * 100
    # Prepare the summary report content
    filtered_virus_taxa_summary = "\n".join([f"{realm}: {count} ({percentage:.1f}%)" for realm, count, percentage in zip(filtered_virus_taxa_counts.index, filtered_virus_taxa_counts.values, filtered_virus_taxa_percentages.values)])

    # Prepare the summary report content for top classes
    filtered_virus_taxa_levels = filtered_quality_summary_concatenated_df['taxonomy'].str.split(';').apply(lambda levels: levels + [''] * (5 - len(levels)))
    filtered_virus_fifth_level_classes = filtered_virus_taxa_levels.apply(lambda x: x[4])
    filtered_virus_filtered_classes = filtered_virus_fifth_level_classes[(filtered_virus_fifth_level_classes != '') & (filtered_virus_fifth_level_classes != 'Unclassified')]
    filtered_virus_class_counts = filtered_virus_filtered_classes.value_counts()
    filtered_virus_top_10_classes = filtered_virus_class_counts.nlargest(10)
    filtered_virus_top_classes_summary = "\n".join([f"{cls}: {count}" for cls, count in filtered_virus_top_10_classes.items()])

    # Calculate summary statistics after both filtration and clustering
    num_vOTUs = len(representative_filtered_quality_summary_concatenated_df)
    num_provOTUs = (representative_filtered_quality_summary_concatenated_df['provirus'].str.contains('Yes', case=False)).sum()
    mean_vOTUs_length = representative_filtered_quality_summary_concatenated_df['virus_length'].mean()
    max_vOTUs_length = representative_filtered_quality_summary_concatenated_df['virus_length'].max()
    min_vOTUs_length = representative_filtered_quality_summary_concatenated_df['virus_length'].min()
    vOTUs_quality_counts = representative_filtered_quality_summary_concatenated_df['checkv_quality'].value_counts()

    # Count occurrences of each realm in the second level of the taxonomy column
    vOTUs_taxa_levels = representative_filtered_quality_summary_concatenated_df['taxonomy'].str.split(';')
    vOTUs_second_level_taxa = vOTUs_taxa_levels.apply(lambda x: x[1] if len(x) > 1 else 'Unclassified/others')
    vOTUs_taxa_counts = vOTUs_second_level_taxa.value_counts()
    # Calculate the percentage of each realm
    vOTUs_taxa_percentages = (vOTUs_taxa_counts / num_vOTUs) * 100
    # Prepare the summary report content
    vOTUs_taxa_summary = "\n".join([f"{realm}: {count} ({percentage:.1f}%)" for realm, count, percentage in zip(vOTUs_taxa_counts.index, vOTUs_taxa_counts.values, vOTUs_taxa_percentages.values)])

    # Prepare the summary report content for top classes
    vOTUs_taxa_levels = representative_filtered_quality_summary_concatenated_df['taxonomy'].str.split(';').apply(lambda levels: levels + [''] * (5 - len(levels)))
    vOTUs_fifth_level_classes = vOTUs_taxa_levels.apply(lambda x: x[4])
    vOTUs_filtered_classes = vOTUs_fifth_level_classes[(vOTUs_fifth_level_classes != '') & (vOTUs_fifth_level_classes != 'Unclassified')]
    vOTUs_class_counts = vOTUs_filtered_classes.value_counts()
    vOTUs_top_10_classes = vOTUs_class_counts.nlargest(10)
    vOTUs_top_classes_summary = "\n".join([f"{cls}: {count}" for cls, count in vOTUs_top_10_classes.items()])

    # Define the path for the new summary report for Module 03
    summary_report_path_module_03 = os.path.join(args["input"], '03_CLUSTERING', 'MVP_03_Summary_Report.txt')

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args["input"],
        '--metadata': args["metadata"],
        '--min_ani': args["min_ani"],
        '--min_tcov': args["min_tcov"],
        '--min_qcov': args["min_qcov"],
        '--threads': args["threads"]}

    # Write a summary line with script arguments and their default values
    summary_line = "03_do_clustering.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    start_time = f"\nStart Time: {mvp_formatted_start_time}\n"
    end_time = f"End time: {mvp_formatted_end_time}\n"
    duration = f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n"

    # Prepare the content for Module 03 summary report
    module_03_summary_content = f"""****************************************************************************
******************               MODULE 03                ******************
****************************************************************************

{summary_line}
{start_time}
{end_time}
{duration}

Summary Report before both filtration and clustering 
--------------------------
Number of Viruses: {num_unfiltered_viruses}, including {num_unfiltered_proviruses} proviruses
Mean Length: {mean_unfiltered_virus_length:.2f}
Max Length: {max_unfiltered_virus_length}
Min Length: {min_unfiltered_virus_length}

Unfiltered Viruses CheckV quality summary:
Low Quality: {unfiltered_virus_quality_counts.get('Low-quality', 0)}
Medium Quality: {unfiltered_virus_quality_counts.get('Medium-quality', 0)}
High Quality: {unfiltered_virus_quality_counts.get('High-quality', 0)}
Complete: {unfiltered_virus_quality_counts.get('Complete', 0)}
Not determined: {unfiltered_virus_quality_counts.get('Not-determined', 0)}

Unfiltered Viruses taxonomy summary:
{unfiltered_virus_taxa_summary}

Top 10 classes of Unfiltered Viruses:
{unfiltered_virus_top_classes_summary}

Summary Report after filtration and before clustering
--------------------------
Number of Viruses: {num_filtered_virus}, including {num_filtered_provirus} proviruses
Mean Length: {mean_filtered_virus_length:.2f}
Max Length: {max_filtered_virus_length}
Min Length: {min_filtered_virus_length}

Filtered Viruses CheckV quality summary:
Low Quality: {filtered_virus_quality_counts.get('Low-quality', 0)}
Medium Quality: {filtered_virus_quality_counts.get('Medium-quality', 0)}
High Quality: {filtered_virus_quality_counts.get('High-quality', 0)}
Complete: {filtered_virus_quality_counts.get('Complete', 0)}
Not determined: {filtered_virus_quality_counts.get('Not-determined', 0)}

Filtered Viruses taxonomy summary:
{filtered_virus_taxa_summary}

Top 10 classes of Filtered Viruses:
{filtered_virus_top_classes_summary}

Summary Report after both filtration and clustering
--------------------------
Number of vOTUs: {num_vOTUs}, including {num_provOTUs} proviruses
Mean Length: {mean_vOTUs_length:.2f}
Max Length: {max_vOTUs_length}
Min Length: {min_vOTUs_length}

vOTUs CheckV quality summary:
Low Quality: {vOTUs_quality_counts.get('Low-quality', 0)}
Medium Quality: {vOTUs_quality_counts.get('Medium-quality', 0)}
High Quality: {vOTUs_quality_counts.get('High-quality', 0)}
Complete: {vOTUs_quality_counts.get('Complete', 0)}
Not determined: {vOTUs_quality_counts.get('Not-determined', 0)}

vOTUs taxonomy summary:
{vOTUs_taxa_summary}

Top 10 classes of vOTUs:
{vOTUs_top_classes_summary}

"""

    # Write the content to the new summary report for Module 03
    with open(summary_report_path_module_03, 'w') as module_03_summary_report:
        module_03_summary_report.write(module_03_summary_content)

def main(args):
    # Capture start time
    mvp_start_time = datetime.datetime.now()
    mvp_formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {mvp_formatted_start_time}") 

    # read metadata file and map FASTQ files
    metadata = pd.read_csv(args["metadata"], sep='\t')

    unfiltered_quality_summary_concatenated_df, filtered_quality_summary_concatenated_df = generate_output_tables(args, rename_sequence)

    output_file, contig_ids_to_keep, blastn_output, ani_output, cluster_output, representative_filtered_quality_summary_concatenated_df = clustering(args, filtered_quality_summary_concatenated_df)

    representative_fasta_file = create_representative_fasta_file(args, output_file, contig_ids_to_keep)

    clean_folder(args, blastn_output, ani_output, cluster_output)

    generate_index(args, representative_fasta_file)
    
    protein_file = generate_protein_files(args, metadata, output_file, representative_fasta_file)

    generate_genomad_annotation_tables(args, protein_file, filtered_quality_summary_concatenated_df, representative_filtered_quality_summary_concatenated_df)

    # Capture end time
    mvp_end_time = datetime.datetime.now()
    mvp_formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    mvp_duration = mvp_end_time - mvp_start_time

    generate_summary_report(args, unfiltered_quality_summary_concatenated_df, filtered_quality_summary_concatenated_df, representative_filtered_quality_summary_concatenated_df, mvp_formatted_start_time, mvp_formatted_end_time, mvp_duration)

    message1 = f"\n\033[1mModule 03 finished: virus genome clustering based on pairwise {args['min_ani']}% ANI done!\033[0m"
    message2 = f"Output files (clustering TXT files and sequence FASTA files) saved in {args['input']}/03_CLUSTERING directory."
    message3 = f"MVP_03_summary_report.txt saved in {args['input']}03_CLUSTERING."
    message4 = f"Reference database from representative viral sequence FASTA file has been generated in {args['input']}/04_READ_MAPPING.\n"
    message4 = f"Representative viral sequence FASTA file and reference will be used as inputs for Module 04 (read mapping)."
    message5 = f"Viral protein FASTA files has been generated in {args['input']}/06_FUNCTIONAL_ANNOTATION directory and will be used as inputs for Module 06 (functional annotation)."
    message6 = "\n\033[1mYou can now proceed to the next step of the MVP script: Module 04!\033[0m\n"
    line_of_stars = '*' * len(message2)
    print()
    print(line_of_stars)
    print(message1)
    print(f"\nStart Time: {mvp_formatted_start_time}") 
    print(f"nd Time: {mvp_formatted_end_time}")
    print(f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(message4)
    print(message5)
    print(message6)
    print(line_of_stars)
    print()

    print("Please don't forget to cite MVP and the following software used by this module:")
    print("- Nayfach, S., Camargo, A.P., Schulz, F. et al. CheckV assesses the quality and completeness of metagenome-assembled viral genomes. Nat Biotechnol 39, 578–585 (2021). https://doi.org/10.1038/s41587-020-00774-7\n")
    print("- Langmead B, Salzberg S. Fast gapped-read alignment with Bowtie 2. Nature Methods. 2012, 9:357-359\n")
    print("- Li H. New strategies to improve minimap2 alignment accuracy. Bioinformatics. 2021, 37(23):4572–4574. https://academic.oup.com/bioinformatics/article/37/23/4572/6384570?login=true\n")


#!/usr/bin/env python3

import argparse
import os
import glob
import pandas as pd
from functools import reduce
import numpy as np
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_05_create_vOTU_table")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument(
        '--covered_fraction',
        nargs='+',
        type=float,
        default=[0.1, 0.5, 0.9],
        help='minimum covered fraction for filtering Default: 0.1 0.5 0.9')
    parser.add_argument(
        "--normalization",
        type=str,
        default='RPKM',
        choices=['RPKM', 'FPKM'],
        help="Metrics to normalize",
    )
    parser.add_argument(
        "--filtration",
        type=str,
        default='conservative',
        choices=['relaxed', 'conservative'],
        help="Filtration level (relaxed or conservative). Default = conservative",
    )
    parser.add_argument(
        "--viral_min_genes",
        type=int,
        default=1,
        help="the minimum number of viral genes required to consider a row as a virus prediction (default = 1)",
    )
    parser.add_argument(
        "--host_viral_genes_ratio",
        type=int,
        default=1,
        help="the maximum ratio of host genes to viral genes required to consider a row as a virus prediction (default = 1",
    )

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()

def merge_coverm_tables(args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Merging all CoverM tables generated in Module 04 (read mapping)...\033[0m")
    # Combine all CoverM CSV files into one
    read_mapping_dir = os.path.join(args['input'], '04_READ_MAPPING', '*', '*_CoverM.tsv')
    all_coverm_csv = glob.glob(read_mapping_dir, recursive=True)
    dfs = []

    for filename in all_coverm_csv:
        df = pd.read_csv(filename, sep='\t', usecols=['Sample', 'Contig', 'Covered Fraction', 'Length', args['normalization']])
        df = df.drop_duplicates(subset=['Sample', 'Contig', 'Length'], keep='last')
        df = df.pivot(index=['Contig', 'Length'], columns='Sample')
        df.columns = ['_'.join(str(col) for col in cols).strip() for cols in df.columns.values]
        dfs.append(df)

    df_coverm_merged = reduce(lambda left, right: pd.merge(left, right, on=['Contig', 'Length'], how='outer'), dfs).reset_index()

    # Read the quality summary TXT file
    quality_summary_concatenated_df_path = os.path.join(args['input'], '03_CLUSTERING', 'MVP_03_All_Sample_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    merged_quality_summary_concatenated_df = pd.read_csv(quality_summary_concatenated_df_path, sep='\t')

    # Merge the CoverM and TSV dataframes
    df_merged = pd.merge(merged_quality_summary_concatenated_df, df_coverm_merged, left_on='virus_id', right_on='Contig', how='right')
    df_merged = df_merged.loc[:, ~df_merged.columns.duplicated(keep='last')]
    df_merged['completeness'].fillna(0, inplace=True)  # When no completeness, set it at 0 so that the sequence does not get filtered out

    # Remove columns that are in the columns_to_exclude list
    columns_to_exclude = ['Contig', 'Length']
    df_merged = df_merged.drop(columns=columns_to_exclude)

    # Save the merged DataFrame as a CSV file in the output folder
    output_filtered_file_path = os.path.join(args['input'], '05_VOTU_TABLES', f'MVP_05_All_Sample_Filtered_{args["filtration"]}_Representative_Virus_Proviruses_vOTU_{args["normalization"]}_Table.tsv')
    os.makedirs(os.path.dirname(output_filtered_file_path), exist_ok=True)
    df_merged.to_csv(output_filtered_file_path, sep='\t', index=False)

    return df_merged, output_filtered_file_path

def filter_coverm_table(df_merged, output_filtered_file_path, args):
    if args['filtration'] == 'conservative':
        print(f"\n\033[1m{step_counter.print_main_step()}: Filtering the merged CoverM table using the conservative mode...\033[0m")
    else:
        print(f"\n\033[1m{step_counter.print_main_step()}: Filtering the merged CoverM table using the relaxed mode...\033[0m")
    # Apply relaxed filtration criteria
    df_filtered = df_merged.loc[(df_merged['viral_genes'] >= args['viral_min_genes'])]
    df_filtered = df_filtered.loc[df_filtered['host_genes']/df_filtered['viral_genes'] <= args['host_viral_genes_ratio']]

    # Apply the new filtration logic if args.filtration is conservative
    if args['filtration'] == 'conservative':
        df_filtered = df_filtered.loc[
            (((df_filtered['completeness_method'].str.contains('AAI-based')) | (df_filtered['completeness_method'].str.contains('DTR'))) &
             (df_filtered['checkv_quality'].isin(['High-quality', 'Medium-quality']))) | (df_filtered['virus_length'] > 5000)]
    
    # Save the filtered DataFrame as a CSV file in the output folder
    df_filtered.to_csv(output_filtered_file_path, sep='\t', index=False)

    return df_filtered

# Define a function to filter and update RPKM columns based on covered fraction
def filter_and_update_RPKM(row, cf, covered_fraction_cols, TPM_cols):
    for sample, cf_col in enumerate(covered_fraction_cols):
        cf_val = row[cf_col]
        if cf_val < cf:
            tpm_col = TPM_cols[sample]
            row[tpm_col] = 0
    return row

def filter_by_coverage_tables(df_filtered, args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Filtering the merged CoverM table using the horizontal coverage and generating a set of final coverage tables...\033[0m")

    df_filtered_list = []

    for cf in args['covered_fraction']:
        # Make a copy of df_filtered so that we don't modify it directly
        df_temp = df_filtered.copy()
        covered_fraction_cols = df_temp.filter(like='Covered Fraction').columns
        TPM_cols = df_temp.filter(like=args['normalization']).columns

        # Apply the filter_and_update_RPKM function to each row
        df_temp = df_temp.apply(filter_and_update_RPKM, args=(cf, covered_fraction_cols, TPM_cols), axis=1)
        
        # Filter out rows where all TPM values are 0
        df_temp = df_temp[~(df_temp[TPM_cols] == 0).all(axis=1)]
        df_temp = df_temp.loc[:, ~df_temp.columns.str.contains('Covered Fraction')]
        df_filtered_list.append(df_temp)

    # Save the filtered dataframes
    for i, cf in enumerate(args['covered_fraction']):
        output_file = os.path.join(args['input'], '05_VOTU_TABLES', f'MVP_05_All_Sample_Filtered_{args["filtration"]}_HC_{cf}_Representative_Virus_Proviruses_vOTU_{args["normalization"]}_Table.tsv')
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        df_filtered_list[i].to_csv(output_file, sep='\t', index=False)

def generate_summary_report(df_filtered, args, formatted_start_time, formatted_end_time, duration):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the summary report for Module 05...\033[0m")
    # Calculate summary statistics
    num_vOTUs = len(df_filtered)
    num_proviruses = (df_filtered['provirus'].str.contains('Yes', case=False)).sum()
    mean_length = df_filtered['virus_length'].mean()
    max_length = df_filtered['virus_length'].max()
    min_length = df_filtered['virus_length'].min()
    quality_counts = df_filtered['checkv_quality'].value_counts()

    # Add columns and counts for RPKM statistics
    rpkm_columns = [col for col in df_filtered.columns if col.startswith('RPKM_')]
    rpkm_virus_counts = [(col, (df_filtered[col] > 0).sum()) for col in rpkm_columns]
    # Sort RPKM counts in descending order
    rpkm_virus_counts.sort(key=lambda x: x[1], reverse=True)

    # Count occurrences of each realm in the second level of the taxonomy column
    taxa_levels = df_filtered['taxonomy'].str.split(';')
    second_level_taxa = taxa_levels.apply(lambda x: x[1] if len(x) > 1 else 'Unclassified/others')
    taxa_counts = second_level_taxa.value_counts()
    # Calculate the percentage of each realm
    taxa_percentages = (taxa_counts / num_vOTUs) * 100
    # Prepare the summary report content
    taxa_summary = "\n".join([f"{realm}: {count} ({percentage:.1f}%)" for realm, count, percentage in zip(taxa_counts.index, taxa_counts.values, taxa_percentages.values)])

    # Prepare the summary report content for top classes
    taxa_levels = df_filtered['taxonomy'].str.split(';').apply(lambda levels: levels + [''] * (5 - len(levels)))
    fifth_level_classes = taxa_levels.apply(lambda x: x[4])
    filtered_classes = fifth_level_classes[(fifth_level_classes != '') & (fifth_level_classes != 'Unclassified')]
    class_counts = filtered_classes.value_counts()
    top_10_classes = class_counts.nlargest(10)
    top_classes_summary = "\n".join([f"{cls}: {count}" for cls, count in top_10_classes.items()])

    # Define the additional lines you want to add to the summary report
    module_05_header = """****************************************************************************
******************               MODULE 05                ******************
****************************************************************************
"""

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args['input'],
        '--metadata': args['metadata'],
        '--covered_fraction': args['covered_fraction'],
        '--normalization': args['normalization'],
        '--filtration': args['filtration'],
        '--viral-min-genes': args['viral_min_genes'],
        '--host-viral-genes-ratio': args['host_viral_genes_ratio']
    }

    # Write a summary line with script arguments and their default values
    summary_line = "05_create_votu_table.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    complete_summary_content = f"""
    \nSummary Report after both clustering and filtration
    --------------------------
    Number of vOTUs: {num_vOTUs}, including {num_proviruses} proviruses
    Mean vOTUs Length: {mean_length:.2f}
    Max vOTUs Length: {max_length}
    Min vOTUs Length: {min_length}

    vOTUs CheckV quality summary:
    Low Quality: {quality_counts.get('Low-quality', 0)}
    Medium Quality: {quality_counts.get('Medium-quality', 0)}
    High Quality: {quality_counts.get('High-quality', 0)}
    Complete: {quality_counts.get('Complete', 0)}
    Not determined: {quality_counts.get('Not-determined', 0)}

    vOTUs taxonomy summary:
    {taxa_summary}

    vOTUs Top 10 classes:
    {top_classes_summary}
    """

    summary_report_path_module_05 = os.path.join(args['input'], '05_VOTU_TABLES', 'MVP_05_Summary_Report.txt')

    # Write the combined content to the new summary report for Module 02
    with open(summary_report_path_module_05, 'w') as module_05_summary_report:
        module_05_summary_report.write(module_05_header + summary_line)
        module_05_summary_report.write(f"\n")
        module_05_summary_report.write(f"\nStart Time: {formatted_start_time}\n")
        module_05_summary_report.write(f"End time: {formatted_end_time}\n")
        module_05_summary_report.write(f"Running Time: {duration.total_seconds():.2f} seconds\n")
        module_05_summary_report.write(complete_summary_content)

def main (args):
    # Capture start time
    mvp_start_time = datetime.datetime.now()
    mvp_formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {mvp_formatted_start_time}") 

    df_merged, output_filtered_file_path = merge_coverm_tables(args)

    df_filtered = filter_coverm_table(df_merged, output_filtered_file_path, args)

    filter_by_coverage_tables(df_filtered, args)

    # Capture end time
    mvp_end_time = datetime.datetime.now()
    mvp_formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    mvp_duration = mvp_end_time - mvp_start_time

    generate_summary_report(df_filtered, args, mvp_formatted_start_time, mvp_formatted_end_time, mvp_duration)

    message1 = "\033[1mModule 05 finished: vOTU tables and summary report generated!\033[0m\n"
    message2 = f"Output vOTU tables and summary report file have been saved in the {args['input']}/05_VOTU_TABLES directory."
    message3 = "\n\033[1mWe hope you enjoyed using MVP script, and you can now explore your data!\033[0m"
    line_of_stars = '*' * len(message2)
    print()
    print(line_of_stars)
    print(message1)
    print(f"Start Time: {mvp_formatted_start_time}") 
    print(f"End Time: {mvp_formatted_end_time}")
    print(f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(line_of_stars)
    print()

    print("Please don't forget to cite the MVP script!\n")


import os
import argparse
import subprocess
import pandas as pd
import shutil
import csv
from Bio import SeqIO
import re
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_06_do_functional_annotation")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument('--fasta_files', 
        type=str, 
        default='representative', 
        choices=['representative', 'all'], 
        help='Sequence and protein FASTA files (representative or all sequences) to use for functional annotation. Default = representative',
    )
    parser.add_argument('--delete_files', 
        action='store_true', 
        help='flag to delete unwanted files',
    )
    parser.add_argument('--PHROGS_evalue', 
        type=int, 
        default=0.01, 
        help='Significance e-value of match between target sequences and query (default = 0.01)',
    )
    parser.add_argument('--PHROGS_score', 
        type=int, 
        default=50, 
        help='Significant score of match between target sequences and query (default = 50)',
    )
    parser.add_argument('--PFAM_evalue', 
        type=int, 
        default=0.01, 
        help='Significance e-value of match between target sequences and query (default = 0.01)',
    )
    parser.add_argument('--PFAM_score', 
        type=int, 
        default=50, 
        help='Significant score of match between target sequences and query (default = 50)',
    )
    parser.add_argument('--ADS', 
        action='store_true', 
        help='Include this flag to search ADS profiles.',
    )
    parser.add_argument('--ADS_evalue', 
        type=int, 
        default=0.01, 
        help='Significance e-value of match between target sequences and query (default = 0.01)',
    )
    parser.add_argument('--ADS_score', 
        type=int, 
        default=60, 
        help='Significant score of match between target sequences and query (default = 60)',
    )
    parser.add_argument('--RdRP', 
        action='store_true', 
        help='Include this flag to create the 07_RDRP_PHYLOGENY folder and search RdRP profiles.',
    )
    parser.add_argument('--RdRP_evalue', 
        type=int, 
        default=0.001, 
        help='Significance e-value of match between target sequences and query (default = 0.01)',
    )
    parser.add_argument('--RdRP_score', 
        type=int, 
        default=50, 
        help='Significant score of match between target sequences and query (default = 50)',
    )
    parser.add_argument('--DRAM', 
        action='store_true', 
        help='Include this flag to create a file to be process through DRAM-v.',
    )
    parser.add_argument('--force_prodigal', 
        action='store_true', 
        help='Force execution of protein prediction by Prodigal.',
    )
    parser.add_argument('--force_PHROGS', 
        action='store_true', 
        help='Force PHROGS annotation.',
    )
    parser.add_argument('--force_PFAM', 
        action='store_true', 
        help='Force PFAM annotation.',
    )
    parser.add_argument('--force_ADS', 
        action='store_true', 
        help='Force ADS annotation.',
    )
    parser.add_argument('--force_RdRP', 
        action='store_true', 
        help='Force RdRP annotation.',
    )
    parser.add_argument('--force_outputs', 
        action='store_true', 
        help='Force creation of final annotation table even though it exists.',
    )
    parser.add_argument('--threads', 
        type=int, 
        default=1, 
        help='Number of threads to use (default = 1)',
    )

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()

def set_up_paths(args):
    # Construct the path to the phrogs file
    script_directory = os.path.dirname(os.path.abspath(__file__))

    if args['fasta_files'] == 'all':
        proteins_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_All_Sample_Filtered_Relaxed_Virus_Provirus_Protein_Sequences.faa')
        sequence_file = os.path.join(args['input'], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences.fna')
        genomad_annotation = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv')
    else:
        proteins_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_All_Sample_Filtered_Relaxed_Representative_Virus_Provirus_Protein_Sequences.faa')
        sequence_file = os.path.join(args['input'], '03_CLUSTERING', 'MVP_03_All_Sample_Filtered_Relaxed_Representative_Virus_Provirus_Sequences.fna')
        genomad_annotation = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Representative_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv')

    if args['RdRP'] and args['ADS']:
        final_annotation_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(genomad_annotation))[0] + '_PHROGS_PFAM_ADS_RDRP_Filtered.tsv')
    elif args['RdRP']:
        final_annotation_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(genomad_annotation))[0] + '_PHROGS_PFAM_RDRP_Filtered.tsv')
    elif args['ADS']:
        final_annotation_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(genomad_annotation))[0] + '_PHROGS_PFAM_ADS_Filtered.tsv')
    else:
        final_annotation_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(genomad_annotation))[0] + '_PHROGS_PFAM_Filtered.tsv')

    phrogs_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_PHROGS.tsv'))
    pfam_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_PFAM.tsv'))
    hmmscan_ADS_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_hmmscan_ADS.tsv'))
    diamond_ADS_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_diamond_ADS.tsv'))
    parsed_hmmscan_ADS_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_hmmscan_ADS_parsed.tsv'))
    parsed_diamond_ADS_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_diamond_ADS_parsed.tsv'))
    unfiltered_parsed_ADS_output_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.basename(genomad_annotation).replace('_GENOMAD.tsv', '_Parsed_ADS.tsv'))
    temp_directory = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'tmp')

    return script_directory, proteins_file, sequence_file, final_annotation_output_file, phrogs_output_file, temp_directory, pfam_output_file, genomad_annotation, hmmscan_ADS_output_file, diamond_ADS_output_file, parsed_hmmscan_ADS_output_file, parsed_diamond_ADS_output_file, unfiltered_parsed_ADS_output_file
    
def predict_protein(args, proteins_file, sequence_file):
    # Check if the output proteins file exists
    if not os.path.exists(proteins_file) or args["force_prodigal"]:
        # The file doesn't exist, so run Prodigal to predict proteins
        print(f"\n\033[1m{step_counter.print_main_step()}: Running Prodigal to predict proteins because protein file doesn't exist or --force_prodigal provided...\033[0m")
        subprocess.run(['prodigal', '-i', sequence_file, '-a', proteins_file, '--threads', str(args['threads'])])
    else:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping Prodigal to predict proteins because {proteins_file} already exists...\033[0m")

def generate_target_sequence(proteins_file):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating a target sequence file as an input for the following functional annotation steps...\033[0m")

    target_sequences = os.path.splitext(proteins_file)[0] + '_target_sequences'
    subprocess.run(['mmseqs', 'createdb', proteins_file, target_sequences])

    return target_sequences

def custom_sort_key(viral_gene_id):
    # Sort rows by Contig_name and Gene_number using the custom sorting key
    parts = viral_gene_id.split('_')
    Contig_name = '_'.join(parts[:-1])
    gene_number = int(parts[-1]) if len(parts) > 1 else 0
    return Contig_name, gene_number

def phrogs_annotation(args, proteins_file, target_sequences,  phrogs_output_file, temp_directory):
    # Load the PHROGS database files
    phrogs_db = os.path.join(args["input"], '00_DATABASES', 'PhrogDB_v14', 'phrogs_prof')
    phrogs_index = os.path.join(args["input"], '00_DATABASES', 'PhrogDB_v14', 'PHROGS_index_filter.csv')

    # Run mmseqs search against PHROGS database
    phrogs = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(proteins_file))[0] + '_mmseqs2_PHROGS')
    if os.path.exists(phrogs_output_file) and not args["force_PHROGS"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping PHROGS annotation, {phrogs_output_file} file already exists and --force_PHROGS not provided...\033[0m")
    else:
        if os.path.exists(phrogs) and not args["force_PHROGS"]:
            print(f"\n\033[1m{step_counter.print_sub_step()}: Skipping mmseqs search for {target_sequences}, {phrogs} file already exists and --force_PHROGS not provided......\033[0m")
        else:
            print(f"\n\033[1m{step_counter.print_sub_step()}: Annotating proteins against PHROGS database...\033[0m")
            subprocess.run(['mmseqs', 'search', phrogs_db, target_sequences, phrogs, temp_directory, '-s', '7', '--threads', str(args['threads'])])

        print(f"\n\033[1m{step_counter.print_sub_step()}: Creating PHROGS annotation output table...\033[0m")
        subprocess.run(['mmseqs', 'createtsv', phrogs_db, target_sequences, phrogs, phrogs_output_file, '--threads', str(args['threads'])])

        phrogs_output_file_df = pd.read_csv(phrogs_output_file, sep='\t', encoding_errors='ignore', names=['#phrog', 'Viral_gene_ID', 'PHROGS_Score', 'PHROGS_seqIdentity', 'PHROGS_evalue', 'qStart', 'qEnd', 'qLen', 'tStart', 'tEnd', 'tLen'])
        phrogs_output_file_df['#phrog'] = phrogs_output_file_df['#phrog'].str.replace('.fma', '')
        phrogs_output_file_df[['Contig_name', 'Gene_number']] = phrogs_output_file_df['Viral_gene_ID'].apply(custom_sort_key).apply(pd.Series)
        phrogs_output_file_df.insert(2, 'Contig_name', phrogs_output_file_df.pop('Contig_name'))
        phrogs_output_file_df.insert(3, 'Gene_number', phrogs_output_file_df.pop('Gene_number'))
        phrogs_output_file_df = phrogs_output_file_df.sort_values(by=['Contig_name', 'Gene_number'], ascending=[True, True])

        # Merge phrogs_output_csv_file and phrogs_index based on the '#phrog' column
        phrogs_index_df = pd.read_csv(phrogs_index)
        annotation_phrogs_output_file_df = pd.merge(phrogs_output_file_df, phrogs_index_df, on='#phrog', how='left').fillna('Unknown')
        annotation_phrogs_output_file_df['PHROGS_Category'] = annotation_phrogs_output_file_df['PHROGS_Category'].str.replace(' function', '')
        annotation_phrogs_output_file_df = annotation_phrogs_output_file_df.drop(columns=annotation_phrogs_output_file_df.columns[0])
        annotation_phrogs_output_file_df.to_csv(phrogs_output_file, sep='\t')

def pfam_annotation(args, proteins_file, target_sequences, pfam_output_file, temp_directory):
    # Load the PFAM database files
    pfam_db = os.path.join(args["input"], '00_DATABASES', 'Pfam_A_DB', 'pfam')
    pfam_index = os.path.join(args["input"], '00_DATABASES', 'Pfam_A_DB', 'PFAM_Index.tsv')
    
    # Run mmseqs search against PFAM database
    pfam = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', os.path.splitext(os.path.basename(proteins_file))[0] + '_mmseqs2_PFAM')
    if os.path.exists(pfam_output_file) and not args["force_PFAM"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping PFAM annotation, {pfam_output_file} file already exists and --force_PFAM not provided...\033[0m")
    else:
        if os.path.exists(pfam) and not args["force_PFAM"]:
            print(f"\n\033[1m{step_counter.print_sub_step()}: Skipping mmseqs search for {target_sequences}, {pfam} file already exists and --force_PFAM not provided...\033[0m")
        else:
            print(f"\n\033[1m{step_counter.print_sub_step()}: Annotating proteins against Pfam database...\033[0m")
            subprocess.run(['mmseqs', 'search', pfam_db, target_sequences, pfam, temp_directory, '-s', '7', '--threads', str(args['threads'])])

        print(f"\n\033[1m{step_counter.print_sub_step()}: Creating Pfam annotation output table...\033[0m")
        subprocess.run(['mmseqs', 'createtsv', pfam_db, target_sequences, pfam, pfam_output_file, '--threads', str(args['threads'])])

        # Read pfam annotation table to merge it with pfam index
        pfam_output_file_df = pd.read_csv(pfam_output_file, sep='\t', encoding_errors='ignore', names=['PFAM_Accession_Number', 'Viral_gene_ID', 'PFAM_Score', 'PFAM_seqIdentity', 'PFAM_evalue', 'qStart', 'qEnd', 'qLen', 'tStart', 'tEnd', 'tLen'])
        pfam_output_file_df['PFAM_Accession_Number'] = pfam_output_file_df['PFAM_Accession_Number'].str.split('.').str[0]
        pfam_index_df = pd.read_csv(pfam_index, sep='\t', encoding_errors='ignore')
        annotation_pfam_output_file_df = pd.merge(pfam_output_file_df, pfam_index_df, left_on='PFAM_Accession_Number', right_on='#pfam', how='left').fillna('Unknown')
        annotation_pfam_output_file_df.to_csv(pfam_output_file, sep='\t')


def merge_phrogs_pfam_tables(args, phrogs_output_file, genomad_annotation, pfam_output_file):
    # Read PHROGS output file
    annotation_phrogs_output_file_df = pd.read_csv(phrogs_output_file, sep='\t', encoding_errors='ignore', usecols=['Viral_gene_ID', 'PHROGS_Score', 'PHROGS_seqIdentity', 'PHROGS_evalue', 'PHROGS_Annotation',
                                                                                                        'PHROGS_Category', 'PHROGS_RefSeq_Annotation', 'PHROGS_Pfam_Annotation',
                                                                                                        'PHROGS_GO_Annotation', 'PHROGS_KO_Annotation'])
    # Filter PHROGS output based on score and evalue
    filter_annotated_phrogs_output_file_df = annotation_phrogs_output_file_df.loc[(annotation_phrogs_output_file_df['PHROGS_Score'] >= args['PHROGS_score']) & (annotation_phrogs_output_file_df['PHROGS_evalue'] <= args['PHROGS_evalue'])]
    # Sort and drop duplicates based on Viral_gene_ID
    filter_annotated_phrogs_output_file_df = filter_annotated_phrogs_output_file_df.sort_values('PHROGS_Score', ascending=False).drop_duplicates('Viral_gene_ID').sort_index()

    # Read Genomad annotation file
    genomad_annotation_df = pd.read_csv(genomad_annotation, sep='\t')
    # Merge Genomad annotation with filtered PHROGS output
    all_genes_filter_annotated_phrogs_output_file_df = pd.merge(genomad_annotation_df, filter_annotated_phrogs_output_file_df, on='Viral_gene_ID', how='left').fillna('NA')

    # Read PFAM output file
    annotation_pfam_output_file_df = pd.read_csv(pfam_output_file, sep='\t', encoding_errors='ignore', usecols=['Viral_gene_ID', 'PFAM_Accession_Number', 'PFAM_Annotation', 'PFAM_Annotation_Short', 'PFAM_Category', 'PFAM_seqIdentity', "PFAM_Score", "PFAM_evalue"])
    # Filter PFAM output based on score and evalue
    PFAM_Index_Filter_df = annotation_pfam_output_file_df.loc[(annotation_pfam_output_file_df['PFAM_Score'] >= args['PFAM_score']) & (annotation_pfam_output_file_df['PFAM_evalue'] <= args['PFAM_evalue'])]
    # Sort and drop duplicates based on Viral_gene_ID
    PFAM_Index_Filter_df = PFAM_Index_Filter_df.sort_values('PFAM_Score', ascending=False).drop_duplicates('Viral_gene_ID').sort_index()
    # Merge Genomad, PHROGS, and PFAM annotations
    PFAM_GENOMAD_PHROGS_df = pd.merge(all_genes_filter_annotated_phrogs_output_file_df, PFAM_Index_Filter_df, on='Viral_gene_ID', how='left').fillna('NA')

    return PFAM_GENOMAD_PHROGS_df

def ADS_annotation(args, script_directory, proteins_file, hmmscan_ADS_output_file, diamond_ADS_output_file, parsed_hmmscan_ADS_output_file, parsed_diamond_ADS_output_file, unfiltered_parsed_ADS_output_file):
    hmm_ADS_db = os.path.abspath(os.path.join(args["input"], '00_DATABASES', 'dbAPIS', 'dbAPIS.hmm'))
    diamond_ADS_db = os.path.join(args["input"], '00_DATABASES', 'dbAPIS', 'APIS_db')
    parse_annotation_result_script_path = os.path.join(script_directory, 'parse_annotation_result.sh')
    ADS_db_index_path = os.path.join(args["input"], '00_DATABASES', 'dbAPIS', 'seed_family_mapping.tsv')

    if os.path.exists(unfiltered_parsed_ADS_output_file) and not args['force_ADS']:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping Anti-Defense System annotation, {unfiltered_parsed_ADS_output_file} file already exists and --force_ADS not provided...\033[0m")
        unfiltered_parsed_ADS_output_df = pd.read_csv(unfiltered_parsed_ADS_output_file, sep='\t', encoding_errors='ignore')
        unfiltered_parsed_ADS_output_df['ADS_score'] = pd.to_numeric(unfiltered_parsed_ADS_output_df['ADS_score'], errors='coerce')
        unfiltered_parsed_ADS_output_df['ADS_evalue'] = pd.to_numeric(unfiltered_parsed_ADS_output_df['ADS_evalue'], errors='coerce')
        filtered_parsed_ADS_output_df = unfiltered_parsed_ADS_output_df.loc[(unfiltered_parsed_ADS_output_df['ADS_score'] >= args['ADS_score']) & (unfiltered_parsed_ADS_output_df['ADS_evalue'] <= args['ADS_evalue'])]
        filtered_parsed_ADS_output_df = filtered_parsed_ADS_output_df.sort_values('ADS_score', ascending=False).drop_duplicates('Viral_gene_ID').sort_index()
    else:
        print(f"\n\033[1m{step_counter.print_sub_step()}: Annotating proteins against Anti-Defense System dbAPIS database using HMMER...\033[0m")
        subprocess.run(['hmmscan', '--domtblout', hmmscan_ADS_output_file, '--noali', hmm_ADS_db, proteins_file], stdout=subprocess.DEVNULL)

        print(f"\n\033[1m{step_counter.print_sub_step()}: Annotating proteins against Anti-Defense System dbAPIS database using DIAMOND...\033[0m")
        subprocess.run(['diamond', 'blastp', '--db', diamond_ADS_db, '-q', proteins_file, '-f', '6', 'qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore', 'qlen', 'slen', '-o', diamond_ADS_output_file, '--max-target-seqs', '10000', '--threads', str(args['threads'])], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        print(f"\n\033[1m{step_counter.print_sub_step()}: Parsing HMMER and DIAMOND annotation tables...\033[0m")
        subprocess.run(['bash', parse_annotation_result_script_path, hmmscan_ADS_output_file, diamond_ADS_output_file, ADS_db_index_path, parsed_hmmscan_ADS_output_file, parsed_diamond_ADS_output_file])

        usecols = ['Viral_gene_ID', 'ADS_Hit_family', 'ADS_Defense_type', 'ADS_Hit_CLAN', 'ADS_Hit_CLAN_Defense_type', 'ADS_evalue', 'ADS_score']

        hmmscan_ADS_output_df = pd.read_csv(parsed_hmmscan_ADS_output_file, sep='\t', encoding_errors='ignore', 
                                            names=['Viral_gene_ID', 'ADS_Query_len', 'ADS_Hit_family', 'ADS_Defense_type', 'ADS_Hit_CLAN', 'ADS_Hit_CLAN_Defense_type', 'ADS_Family_len',
                                                    'ADS_evalue', 'ADS_score', 'ADS_Query_from', 'ADS_Query_to', 'ADS_HMM_from', 'ADS_HMM_to'])
        hmmscan_ADS_output_df = hmmscan_ADS_output_df[usecols]
        
        diamond_ADS_output_df = pd.read_csv(parsed_diamond_ADS_output_file, sep='\t', encoding_errors='ignore', 
                                    names=['Viral_gene_ID', 'ADS_Hit_family', 'ADS_Defense_type', 'ADS_Hit_CLAN', 'ADS_Hit_CLAN_Defense_type', 'ADS_seqid', 'ADS_pident', 'ADS_align_length',
                                        'ADS_evalue', 'ADS_score', 'ADS_qcov', 'ADS_scov'])
        diamond_ADS_output_df = diamond_ADS_output_df[usecols]
        
        unfiltered_parsed_ADS_output_df = pd.concat([hmmscan_ADS_output_df, diamond_ADS_output_df])
        unfiltered_parsed_ADS_output_df.to_csv(unfiltered_parsed_ADS_output_file, sep='\t', index=False)

        # Convert the 'score' column to numeric
        unfiltered_parsed_ADS_output_df['ADS_score'] = pd.to_numeric(unfiltered_parsed_ADS_output_df['ADS_score'], errors='coerce')
        unfiltered_parsed_ADS_output_df['ADS_evalue'] = pd.to_numeric(unfiltered_parsed_ADS_output_df['ADS_evalue'], errors='coerce')

        filtered_parsed_ADS_output_df = unfiltered_parsed_ADS_output_df.loc[(unfiltered_parsed_ADS_output_df['ADS_score'] >= args['ADS_score']) & (unfiltered_parsed_ADS_output_df['ADS_evalue'] <= args['ADS_evalue'])]
        filtered_parsed_ADS_output_df = filtered_parsed_ADS_output_df.sort_values('ADS_score', ascending=False).drop_duplicates('Viral_gene_ID').sort_index()

    return filtered_parsed_ADS_output_df

def RDRP_annotation(args, proteins_file):
    rdrp_phylogeny_dir = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', '06_RDRP_ANNOTATION')
    os.makedirs(rdrp_phylogeny_dir, exist_ok=True)
    # Load the RdRP profile HMM file
    RdRP_profile = os.path.abspath(os.path.join(args["input"], '00_DATABASES', 'RdRP_DB', 'All_RdRP_profiles_Wolf_set.22.hmm'))
    RdRP_Profile_Output = os.path.join(rdrp_phylogeny_dir, '06A_RdRP_Profile_Tab.txt')
    Formatted_RdRP_Profile_Output = os.path.join(rdrp_phylogeny_dir, '06B_Formatted_RdRP_Profile_Tab.tsv')
    Filtered_Formatted_RdRP_Profile_Output = os.path.join(rdrp_phylogeny_dir, '06C_Filtered_Formatted_RdRP_Profile_Tab.tsv')

    if os.path.exists(Filtered_Formatted_RdRP_Profile_Output) and not args["force_RdRP"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping annotating proteins against RdRP HMM profiles, {Filtered_Formatted_RdRP_Profile_Output} file already exists and --force_RdRP not provided...\033[0m")
        Filtered_Formatted_RdRP_Profile_Output_df = pd.read_csv(Filtered_Formatted_RdRP_Profile_Output, sep='\t', encoding_errors='ignore')
    else:
        print(f"\n\033[1m{step_counter.print_sub_step()}: Annotating proteins against RdRP HMM profiles...\033[0m")
        subprocess.run(['hmmsearch', '-o', os.path.join(rdrp_phylogeny_dir, '06A_RdRP_Profile_Output.txt'), '--tblout', RdRP_Profile_Output, RdRP_profile, proteins_file])

        with open(os.path.join(rdrp_phylogeny_dir, '06A_RdRP_Profile_Tab.txt'), 'r') as infile:
            lines = infile.readlines()

        # Remove the first three rows and the last rows containing '#'
        lines = [line for line in lines[3:] if not line.startswith('#')]
        data = [line.strip().split() for line in lines]

        # Filter the data to keep only the desired columns if they exist
        filtered_data = []
        for row in data:
            if len(row) >= 7:
                filtered_data.append([row[0], row[2], row[4], row[5]])

        print(f"\n\033[1m{step_counter.print_sub_step()}: Creating and filtering the RdRP output tables...\033[0m")
        Formatted_RdRP_Profile_Output_df = pd.DataFrame(filtered_data, columns=["Viral_gene_ID", "RDRP_Annotation", "RDRP_evalue", "RDRP_Score"])
        Formatted_RdRP_Profile_Output_df.to_csv(Formatted_RdRP_Profile_Output, sep='\t', index=False)

        # Convert the 'score' column to numeric
        Formatted_RdRP_Profile_Output_df['RDRP_Score'] = pd.to_numeric(Formatted_RdRP_Profile_Output_df['RDRP_Score'], errors='coerce')
        Formatted_RdRP_Profile_Output_df['RDRP_evalue'] = pd.to_numeric(Formatted_RdRP_Profile_Output_df['RDRP_evalue'], errors='coerce')

        Filtered_Formatted_RdRP_Profile_Output_df = Formatted_RdRP_Profile_Output_df.sort_values('RDRP_Score').drop_duplicates('Viral_gene_ID', keep='last')
        Filtered_Formatted_RdRP_Profile_Output_df = Filtered_Formatted_RdRP_Profile_Output_df[Filtered_Formatted_RdRP_Profile_Output_df['RDRP_Score'] >= args['RdRP_score']]
        Filtered_Formatted_RdRP_Profile_Output_df = Filtered_Formatted_RdRP_Profile_Output_df[Filtered_Formatted_RdRP_Profile_Output_df['RDRP_evalue'] <= args['RdRP_evalue']]
        Filtered_Formatted_RdRP_Profile_Output_df.to_csv(Filtered_Formatted_RdRP_Profile_Output, sep='\t', index=False)

    return Filtered_Formatted_RdRP_Profile_Output_df

def merge_final_annotation_tables(args, final_annotation_output_file, PFAM_GENOMAD_PHROGS_df, Filtered_Formatted_RdRP_Profile_Output_df, filtered_parsed_ADS_output_df):
    if args['RdRP'] and args['ADS']:
        print(f"\n\033[1m{step_counter.print_main_step()}: Merging geNomad, PHROGS, PFAM, ADS, and RdRP annotation tables (filtration mode = relaxed)...\033[0m")
        RDRP_GENOMAD_PHROGS_PFAM_df = pd.merge(PFAM_GENOMAD_PHROGS_df, Filtered_Formatted_RdRP_Profile_Output_df[["Viral_gene_ID", "RDRP_Annotation", "RDRP_evalue", "RDRP_Score"]], on='Viral_gene_ID', how='left').fillna('NA')
        RDRP_GENOMAD_PHROGS_PFAM_df.insert(12, 'RDRP_Annotation', RDRP_GENOMAD_PHROGS_PFAM_df.pop('RDRP_Annotation'))
        RDRP_GENOMAD_PHROGS_PFAM_ADS_df = pd.merge(RDRP_GENOMAD_PHROGS_PFAM_df, filtered_parsed_ADS_output_df, on='Viral_gene_ID', how='left').fillna('NA')
        RDRP_GENOMAD_PHROGS_PFAM_ADS_df.insert(13, 'ADS_Hit_family', RDRP_GENOMAD_PHROGS_PFAM_ADS_df.pop('ADS_Hit_family'))
        RDRP_GENOMAD_PHROGS_PFAM_ADS_df.insert(14, 'ADS_Defense_type', RDRP_GENOMAD_PHROGS_PFAM_ADS_df.pop('ADS_Defense_type'))
        desired_column_order = ['Viral_gene_ID', 'Contig_name', 'Gene_number', 'start', 'end', 'length', 'strand', 'GENOMAD_Annotation', 'PHROGS_Annotation', 'PHROGS_Category', 'PFAM_Annotation',	'PFAM_Category', "RDRP_Annotation",
                                'ADS_Hit_family', 'ADS_Defense_type', 'GENOMAD_Annotation_accessions', 'GENOMAD_Score', 'GENOMAD_evalue', 'GENOMAD_gc_content', 'GENOMAD_genetic_code', 'GENOMAD_rbs_motif', 'GENOMAD_marker',
                                'GENOMAD_virus_hallmark', 'GENOMAD_Annotation_amr', 'PHROGS_Score', 'PHROGS_evalue', 'PHROGS_seqIdentity', 'PHROGS_RefSeq_Annotation', 'PHROGS_Pfam_Annotation',
                                'PHROGS_GO_Annotation', 'PHROGS_KO_Annotation', 'PFAM_Accession_Number', 'PFAM_Annotation_Short', 'PFAM_seqIdentity', 'PFAM_Score', 'PFAM_evalue', "RDRP_evalue", "RDRP_Score", 'ADS_Hit_CLAN', 'ADS_Hit_CLAN_Defense_type', 'ADS_evalue', 'ADS_score']
        RDRP_GENOMAD_PHROGS_PFAM_ADS_df = RDRP_GENOMAD_PHROGS_PFAM_ADS_df[desired_column_order]
        RDRP_GENOMAD_PHROGS_PFAM_ADS_df.to_csv(final_annotation_output_file, sep='\t', index=False)
    elif args['RdRP']:
        print(f"\n\033[1m{step_counter.print_main_step()}: Merging geNomad, PHROGS, PFAM, and RdRP annotation tables (filtration mode = relaxed)...\033[0m")
        RDRP_GENOMAD_PHROGS_PFAM_df = pd.merge(PFAM_GENOMAD_PHROGS_df, Filtered_Formatted_RdRP_Profile_Output_df[["Viral_gene_ID", "RDRP_Annotation", "RDRP_evalue", "RDRP_Score"]], on='Viral_gene_ID', how='left').fillna('NA')
        RDRP_GENOMAD_PHROGS_PFAM_df.insert(12, 'RDRP_Annotation', RDRP_GENOMAD_PHROGS_PFAM_df.pop('RDRP_Annotation'))
        RDRP_GENOMAD_PHROGS_PFAM_df.to_csv(final_annotation_output_file, sep='\t', index=False)
    elif args['ADS']:
        print(f"\n\033[1m{step_counter.print_main_step()}: Merging geNomad, PHROGS, PFAM, and ADS annotation tables (filtration mode = relaxed)...\033[0m")
        ADS_GENOMAD_PHROGS_PFAM_df = pd.merge(PFAM_GENOMAD_PHROGS_df, filtered_parsed_ADS_output_df, on='Viral_gene_ID', how='left').fillna('NA')
        desired_column_order = ['Viral_gene_ID', 'Contig_name', 'Gene_number', 'start', 'end', 'length', 'strand', 'GENOMAD_Annotation', 'PHROGS_Annotation', 'PHROGS_Category', 'PFAM_Annotation',	'PFAM_Category',
                                'ADS_Hit_family', 'ADS_Defense_type', 'GENOMAD_Annotation_accessions', 'GENOMAD_Score', 'GENOMAD_evalue', 'GENOMAD_gc_content', 'GENOMAD_genetic_code', 'GENOMAD_rbs_motif', 'GENOMAD_marker',
                                'GENOMAD_virus_hallmark', 'GENOMAD_Annotation_amr', 'PHROGS_Score', 'PHROGS_evalue', 'PHROGS_seqIdentity', 'PHROGS_RefSeq_Annotation', 'PHROGS_Pfam_Annotation',
                                'PHROGS_GO_Annotation', 'PHROGS_KO_Annotation', 'PFAM_Accession_Number', 'PFAM_Annotation_Short', 'PFAM_seqIdentity', 'PFAM_Score', 'PFAM_evalue', 'ADS_Hit_CLAN', 'ADS_Hit_CLAN_Defense_type', 'ADS_evalue', 'ADS_score']
        ADS_GENOMAD_PHROGS_PFAM_df = ADS_GENOMAD_PHROGS_PFAM_df[desired_column_order]
        ADS_GENOMAD_PHROGS_PFAM_df.to_csv(final_annotation_output_file, sep='\t', index=False)
    else:
        print(f"\n\033[1m{step_counter.print_main_step()}: Merging geNomad, PHROGS, and PFAM annotation tables (filtration mode = relaxed)...\033[0m")
        PFAM_GENOMAD_PHROGS_df.to_csv(final_annotation_output_file, sep='\t', index=False)

    print(f"Final annotation table {final_annotation_output_file} created!") 

def conservative_mode_final_table(args, final_annotation_output_file):
    # Define the filtration argument used in Module 05
    summary_report_path_module_05 = os.path.join(args['input'], '05_VOTU_TABLES', 'MVP_05_Summary_Report.txt')
    module_05_summary_content = ""
    if os.path.exists(summary_report_path_module_05):  # We need this test in case users ran module 01 - 02 - 03 then 06 (without 04 / 05)
    
        with open(summary_report_path_module_05, 'r') as module_05_summary_report:
            module_05_summary_content = module_05_summary_report.read()

        target_line = next((line for line in module_05_summary_content.split('\n') if line.startswith('05_create_votu_table.py')), None)
        filtration_option = re.search(r'--filtration (\w+)', target_line).group(1) if target_line else None

        if filtration_option == 'conservative':
            final_relaxed_functional_annotation_df = pd.read_csv(final_annotation_output_file, sep='\t', encoding_errors='ignore')
            filtered_conservative_vOTUs_path = os.path.join(args['input'], '05_VOTU_TABLES', 'MVP_05_All_Sample_Filtered_conservative_Representative_Virus_Proviruses_vOTU_RPKM_Table.tsv')
            filtered_conservative_vOTUs_df = pd.read_csv(filtered_conservative_vOTUs_path, sep='\t', encoding_errors='ignore')
            final_conservative_functional_annotation_df = pd.merge(filtered_conservative_vOTUs_df[["virus_id"]], final_relaxed_functional_annotation_df, how="left", left_on="virus_id", right_on="Contig_name")
            final_conservative_functional_annotation_df = final_conservative_functional_annotation_df.drop(columns=['virus_id'])
            final_conservative_functional_annotation_path = final_annotation_output_file.replace("Relaxed", "Conservative")
            final_conservative_functional_annotation_df.to_csv(final_conservative_functional_annotation_path, sep='\t', index=False)
    else:
        filtration_option = "none"
    
    # Calculate summary statistics
    if filtration_option == 'conservative':
        print(f"\n\033[1m{step_counter.print_main_step()}: Filtering final annotation tables based on conservative mode if used in Module 05...\033[0m")
        final_annotation_output_file_df = pd.read_csv(final_conservative_functional_annotation_path, sep='\t', encoding_errors='ignore')
    else:
        print(f"\n\033[1m{step_counter.print_main_step()}: Filtering final annotation tables based on relaxed mode (used in Module 05)...\033[0m")
        final_annotation_output_file_df = pd.read_csv(final_annotation_output_file, sep='\t', encoding_errors='ignore')

    return final_annotation_output_file_df

def guess_vs_hallmark(vec):
    if vec['GENOMAD_marker'] == "NA" or pd.isna(vec['GENOMAD_marker']):
        return "-"
    t = vec['GENOMAD_marker'].split(".")
    if t[2].startswith("V") or t[2].endswith("V"):
        # We have a virus gene, is it hallmark ? (note - hallmark coded as 0 in VirSorter, 1 in geNomad)
        if vec['GENOMAD_virus_hallmark'] == 1:
            return "0"
        else:
            return "1"
    else:
        return "-"

def get_contig_name(gene_name):
            x = gene_name.split("_")
            return "_".join(x[:-1])

def add_line(vec, outwriter):
    # Convert strand notation in VirSorter format (plus and minus)
    if vec['strand'] == -1:
        strand = "-"
    elif vec["strand"] == 1:
        strand = "+"
    # Convert marker gene name if na
    if vec['VS_hallmark'] == "-":
        vec['GENOMAD_marker'] = "-"
        vec['GENOMAD_Score'] = "-"
        vec['GENOMAD_evalue'] = "-"

    if pd.isna(vec['PFAM_Annotation_Short']):
        PFAM_Annotation_Short = "-"
        PFAM_Score = "-"
        PFAM_evalue = "-"
    else:
        PFAM_Annotation_Short = vec['PFAM_Annotation_Short']
        PFAM_Score = vec['PFAM_Score']
        PFAM_evalue = vec['PFAM_evalue']

    outwriter.writerow(
        [vec['Viral_gene_ID'], vec['start'], vec['end'], vec['length'], strand, vec['GENOMAD_marker'],
            vec['GENOMAD_Score'], vec['GENOMAD_evalue'], vec['VS_hallmark'], PFAM_Annotation_Short, PFAM_Score,
            PFAM_evalue])
    
def write_output(df, outwriter, output_handle):
    # First add a line for the contig itself
    contig = df['contig'].unique()[0]
    n_gene = len(df)
    output_handle.write(f">{contig}|{n_gene}|l\n")  # Note - we falsely claim all sequences are linear, it does not matter really for DRAM-V downstream
    # Now adding one line for each gene, sorted by start
    df.sort_values(['start', 'strand'], ascending=[True, True]).apply(lambda x: add_line(x, outwriter), axis=1)
            
def create_DRAM_inputs(args, genomad_annotation, sequence_file, final_annotation_output_file):
    # Create a file to be processed through DRAM-v if --DRAM flag is provided
    dram_v_folder = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', '06_DRAM_V')
    DRAM_input_file = os.path.join(dram_v_folder, os.path.splitext(os.path.basename(genomad_annotation))[0] + '_DRAM_Input.tsv')
    sequence_output_file = os.path.join(dram_v_folder, os.path.splitext(os.path.basename(sequence_file))[0] + '_DRAM_Input.fna')

    if args['DRAM']:
        print(f"\n\033[1m{step_counter.print_main_step()}: Generating input files for DRAM-v functional annotation analysis...\033[0m")
        os.makedirs(dram_v_folder, exist_ok=True)  # Create the folder if it doesn't exist

        records = list(SeqIO.parse(sequence_file, "fasta"))

        with open(sequence_output_file, "w") as output_handle:
            for record in records:
                # Replace '|' characters with '_' in the sequence header
                modified_header = record.id.replace('|', '_') + '-cat_1'

                # Write a single line with both the original and modified headers
                output_handle.write(f">{modified_header} {record.id}\n{record.seq}\n")

        annotation_df = pd.read_csv(final_annotation_output_file, sep='\t', encoding_errors='ignore',
                                    usecols=['Viral_gene_ID', 'start', 'end', 'length', 'strand',
                                                'GENOMAD_marker',
                                                'GENOMAD_Score', 'GENOMAD_evalue', 'GENOMAD_virus_hallmark',
                                                'PFAM_Annotation', 'PFAM_Annotation_Short', 'PFAM_Accession_Number',
                                                'PFAM_Score', 'PFAM_evalue'])
        # Replace '|' characters with '_' in 'Viral_gene_ID' column
        annotation_df['Viral_gene_ID'] = annotation_df['Viral_gene_ID'].str.replace('|', '_')
        # Add info about hallmark gene in the VirSorter format
        annotation_df['VS_hallmark'] = annotation_df[['GENOMAD_marker', 'GENOMAD_virus_hallmark']].apply(lambda x: guess_vs_hallmark(x), axis=1)
        # Also add info about the original contig
        annotation_df['contig'] = annotation_df['Viral_gene_ID'].apply(lambda x: get_contig_name(x))
        # Prepare the output file
        with open(DRAM_input_file, "w") as output_handle:
            outwriter = csv.writer(output_handle, delimiter='|', quotechar='"', quoting=csv.QUOTE_MINIMAL, lineterminator="\n")
            annotation_df.groupby('contig').apply(lambda x: write_output(x, outwriter, output_handle))

def delete_files(args, temp_directory):
    print(f"\n\033[1m{step_counter.print_main_step()}: Deleting temporary files...\033[0m")
    # Define the directory where you want to remove files
    directory_to_clean = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION')
    # Remove temporary files and folders
    for filename in os.listdir(directory_to_clean):
        if filename.endswith(('.faa', '.tsv', '.txt')):
            continue  # Keep .faa, .tsv, and .txt files
        file_path = os.path.join(directory_to_clean, filename)
        if os.path.isfile(file_path):
            os.remove(file_path)

    # Remove temporary files and folders
    if os.path.exists(temp_directory):
        shutil.rmtree(temp_directory)

def generate_summary_report(args, final_annotation_output_file_df, phrogs_output_file, mvp_start_time, mvp_end_time):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the summary report for Module 06...\033[0m")
    # Calculate the number of genes and hits
    num_genes = len(final_annotation_output_file_df)
    num_hit_genes = len(phrogs_output_file)

    # Count occurrences of each category in the PHROGS_Category column
    category_levels = final_annotation_output_file_df['PHROGS_Category']
    category_counts = category_levels.value_counts()
    # Calculate the percentage of each category
    taxa_percentages = (category_counts / num_genes) * 100
    # Prepare the summary report content
    category_summary = "\n".join([f"{category}: {count} ({percentage:.1f}%)" for category, count, percentage in zip(category_counts.index, category_counts.values, taxa_percentages.values)])

    # Module 06 header
    module_06_header = """****************************************************************************
******************               MODULE 06                ******************
****************************************************************************
"""

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args['input'],
        '--metadata': args['metadata'],
        '--fasta_files': args['fasta_files'],
        '--delete_files': args['delete_files'],
        '--PHROGS_evalue': args['PHROGS_evalue'],
        '--PHROGS_score': args['PHROGS_score'],
        '--PFAM_evalue': args['PFAM_evalue'],
        '--PFAM_score': args['PFAM_score'],
        '--ADS': args['ADS'],
        '--ADS_evalue': args['ADS_evalue'],
        '--ADS_score': args['ADS_score'],
        '--RdRP': args['RdRP'],
        '--RdRP_evalue': args['RdRP_evalue'],
        '--RdRP_score': args['RdRP_score'],
        '--DRAM': args['DRAM'],
        '--force_prodigal': args['force_prodigal'],
        '--force_PHROGS': args['force_PHROGS'],
        '--force_PFAM': args['force_PFAM'],
        '--force_ADS': args['force_ADS'],
        '--force_RdRP': args['force_RdRP'],
        '--force_outputs': args['force_outputs'],
        '--threads': args['threads']
    }

    # Write a summary line with script arguments and their default values
    summary_line = "06_do_functional_annotation.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    # Calculate duration
    duration = mvp_end_time - mvp_start_time
    formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Prepare the complete summary content
    complete_summary_content = f"""
\nSummary Report of the functional annotation after filtration
--------------------------
Number of genes: {num_genes}
Number of hits against PHROGS db: {num_hit_genes}

Gene category summary:
{category_summary}
"""

    summary_report_path_module_06 = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_Summary_Report.txt')

    # Write the combined content to the new summary report for Module 06
    with open(summary_report_path_module_06, 'w') as module_06_summary_report:
        module_06_summary_report.write(module_06_header)
        module_06_summary_report.write(summary_line)
        module_06_summary_report.write(f"\n")
        module_06_summary_report.write(f"\nStart Time: {formatted_start_time}\n")
        module_06_summary_report.write(f"End time: {formatted_end_time}\n")
        module_06_summary_report.write(f"Running Time: {duration.total_seconds():.2f} seconds\n")
        module_06_summary_report.write(complete_summary_content)

def main(args): 
    # Capture start time
    mvp_start_time = datetime.datetime.now()
    mvp_formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {mvp_formatted_start_time}") 

    script_directory, proteins_file, sequence_file, final_annotation_output_file, phrogs_output_file, temp_directory, pfam_output_file, genomad_annotation, hmmscan_ADS_output_file, diamond_ADS_output_file, parsed_hmmscan_ADS_output_file, parsed_diamond_ADS_output_file, unfiltered_parsed_ADS_output_file = set_up_paths(args)

    if not args['force_outputs'] and os.path.exists(final_annotation_output_file):
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping functional annotation steps beacause {final_annotation_output_file}' exists, and --force_outputs argument not provided...\033[0m")
    else:
        predict_protein(args, proteins_file, sequence_file)

        target_sequences = generate_target_sequence(proteins_file)

        phrogs_annotation(args, proteins_file, target_sequences,  phrogs_output_file, temp_directory)

        pfam_annotation(args, proteins_file, target_sequences, pfam_output_file, temp_directory)

        PFAM_GENOMAD_PHROGS_df = merge_phrogs_pfam_tables(args, phrogs_output_file, genomad_annotation, pfam_output_file)

        # ADS prediction
        if args['ADS']:
            filtered_parsed_ADS_output_df = ADS_annotation(args, script_directory, proteins_file, hmmscan_ADS_output_file, diamond_ADS_output_file, parsed_hmmscan_ADS_output_file, parsed_diamond_ADS_output_file, unfiltered_parsed_ADS_output_file)
        else:
            filtered_parsed_ADS_output_df = None

        # Create the "07_RDRP_PHYLOGENY" directory if --RdRP flag is provided
        if args['RdRP']:
            Filtered_Formatted_RdRP_Profile_Output_df = RDRP_annotation(args, proteins_file)
        else:
            Filtered_Formatted_RdRP_Profile_Output_df = None

        merge_final_annotation_tables(args, final_annotation_output_file, PFAM_GENOMAD_PHROGS_df, Filtered_Formatted_RdRP_Profile_Output_df, filtered_parsed_ADS_output_df)

    final_annotation_output_file_df = conservative_mode_final_table(args, final_annotation_output_file)

    create_DRAM_inputs(args, genomad_annotation, sequence_file, final_annotation_output_file)

    if args['delete_files']:
        delete_files(args, temp_directory)

    # Capture end time
    mvp_end_time = datetime.datetime.now()
    mvp_formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    mvp_duration = mvp_end_time - mvp_start_time

    generate_summary_report(args, final_annotation_output_file_df, phrogs_output_file, mvp_start_time, mvp_end_time)

    message1 = f"\033[1mModule 06 finished: functional annotation done and summary report generated!\033[0m"
    message2 = f"Final annotation table {final_annotation_output_file} has been created."
    message3 = f"MVP_06_summary_report.txt saved in {args['input']}/06_FUNCTIONAL_ANNOTATION."
    line_of_stars = '*' * len(message1)
    print()
    print(line_of_stars)
    print(message1)
    print(f"\nStart Time: {mvp_formatted_start_time}") 
    print(f"End Time: {mvp_formatted_end_time}")
    print(f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(line_of_stars)
    print()

    print("Please don't forget to cite MVP and the following software used by this module:")
    print("- Hyatt, D. et al. Prodigal: Prokaryotic Gene Recognition and Translation Initiation Site Identification. BMC Bioinformatics. 2010, 11(1). https://doi.org/10.1186/1471-2105-11-119.\n")
    print("- Mirdita, M. et al. Fast and sensitive taxonomic assignment to metagenomic contigs. Bioinformatics. 2021, 37(18). https://doi.org/10.1093/bioinformatics/btab184\n")
    print("- Potter, S. et al. HMMER Web Server: 2018 Update. Nucleic Acids Research. 2018, 46(1). https://doi.org/10.1093/nar/gky448.\n")



import argparse
import glob
import os
import subprocess
import sys
from os.path import getsize
import pandas as pd
from functools import reduce
from Bio import SeqIO
from importlib_resources import files as resource_files
from tqdm import tqdm
import datetime
import time
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_01_run_genomad_checkv")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument(
        "--sample_group",
        help="Specific sample number(s) to run the script on (can be a comma-separated list: 1,2,6 for example). By default, MVP processes all datasets listed in the metadata file one after the other.",
    )
    parser.add_argument(
        "--skip_modify_assemblies",
        action="store_true",
        default=False,
        help="Modify sequence headers by adding sample name, False by default.",
    )
    parser.add_argument(
        "--min_seq_size",
        type=int,
        default=0,
        help="Minimum sequence size to keep (in base pairs).",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--genomad_relaxed', help='Run geNomad with relaxed filtering.', action='store_true')
    group.add_argument('--genomad_conservative', help='Run geNomad with conservative filtering.', action='store_true')
    parser.add_argument('--genomad_db_path', help='Path to the geNomad database directory.', default='')
    parser.add_argument('--checkv_db_path', help='Path to the CheckV database directory.', default='')
    parser.add_argument('--force_genomad', help='Run geNomad even if output already exists.', action='store_true')
    parser.add_argument('--force_checkv', help='Run CheckV even if output already exists.', action='store_true')
    parser.add_argument(
        '--threads', 
        type=int,
        default=1,
        help='Number of threads to use (default = 1)')

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()
    
# Define a function to find closest matching taxonomy and return Genome type and Host type
def find_closest_match(taxonomy, ictv_df):
    taxonomy_levels = taxonomy.split(';')
    matching_rows = []

    for i in range(len(taxonomy_levels), 0, -1):
        partial_taxonomy = ';'.join(taxonomy_levels[:i])
        match_rows = ictv_df[ictv_df['Family'] == partial_taxonomy]
        if not match_rows.empty:
            matching_rows.extend(match_rows.iterrows())

    if matching_rows:
        return matching_rows[0][1][['Genome type', 'Host type']]
    else:
        return pd.Series({'Genome type': 'Unknown', 'Host type': 'Unknown'})

# Define a function to map taxonomy to genome type
def taxon_to_genome_type(taxonomy, ictv_taxonomy_df):
    genome_host_info = find_closest_match(taxonomy, ictv_taxonomy_df)
    return genome_host_info['Genome type']

# Define a function to map taxonomy to host type
def taxon_to_host_type(taxonomy, ictv_taxonomy_df):
    genome_host_info = find_closest_match(taxonomy, ictv_taxonomy_df)
    return genome_host_info['Host type']

def rename_sequence(sequence_name):
    sequence_name = sequence_name.replace('provirus_', '')
    sequence_name = sequence_name.rsplit('/', 1)[0]
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')
    return sequence_name

def rename_sequence_coord_checkv(virus_id):
    parts = virus_id.split('|')
    checkv = parts[-1]
    genomad = parts[-2]
    sequence_name = ""
    if genomad.startswith("provirus"): 
        tab_genomad = genomad.split('_')[1:3]
        tab_checkv = checkv.split('/')[0].split('-')
        if len(parts)>2:
            parts[0] = "_".join(parts[0:-2])  ## That is here in case there was some pipe in the original contig name
        tab_genomad[0] = str(int(tab_genomad[0]) + int(tab_checkv[0]) - 1)
        tab_genomad[1] = str(int(tab_genomad[0]) + int(tab_checkv[1]) - 1)
        sequence_name = parts[0] + '_' + "-".join(tab_genomad)
    else:
        tab_checkv = checkv.split('/')[0].split('-')
        if len(parts)>1:
            parts[0] = "_".join(parts[0:-1])  ## That is in case there was some pipe in the original contig name
        sequence_name = parts[0] + '_' + "-".join(tab_checkv)
    sequence_name = sequence_name.replace('|', '_').replace('-', '_').replace('/', '_').replace(':', '_')
    return sequence_name

def correct_coordinates(virus_id):
    tab_coords = virus_id.split('_')[-2:] ## So we simply take the last two fields when separating by "_"
    return "-".join(tab_coords)

def modify_assembly_files(input_assembly_file, sample_name, modified_assembly_dir, args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Modifying the assembly files (adding sample name to headers and/or delete short sequences)...\033[0m")

    # Construct the new output file name in the modified assembly directory
    if args["skip_modify_assemblies"]:
        input_for_genomad = input_assembly_file
    else:
        output_assembly_file = os.path.join(modified_assembly_dir, f'{sample_name}_modified.fna')
        # Read sequences, modify headers, filter by size, and write to the new file
        sequences = []
        with open(input_assembly_file, "r") as input_fasta:
            for record in SeqIO.parse(input_fasta, "fasta"):
                # Modify the header by adding the sample name at the beginning
                new_header = f"{sample_name}_{record.id}"
                record.id = new_header
                record.description = new_header

                if len(record.seq) > args["min_seq_size"]:
                    sequences.append(record)

        # Write the modified and filtered sequences to the new file
        with open(output_assembly_file, "w") as output_fasta:
            SeqIO.write(sequences, output_fasta, 'fasta-2line')

        # Use the output_assembly_file for Genomad if it was modified
        input_for_genomad = output_assembly_file

    return input_for_genomad

def run_genomad_first_run(sample_name, input_for_genomad, genomad_output, args, genomad_db_path):
    if os.path.exists(genomad_output) and not args["force_genomad"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping running geNomad for {sample_name}, {genomad_output} folder already exists, and force_genomad not provided...\033[0m")
    else:
        if args["genomad_relaxed"]:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running geNomad (relaxed mode) on {sample_name}...\033[0m")
            subprocess.run(['genomad', 'end-to-end', '--relaxed', '--cleanup', '--splits', '8', '-t', str(args["threads"]), input_for_genomad, genomad_output, genomad_db_path])
        elif args["genomad_conservative"]:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running geNomad (conservative mode) on {sample_name}...\033[0m")
            subprocess.run(['genomad', 'end-to-end', '--conservative', '--cleanup', '--splits', '8', '-t', str(args["threads"]), input_for_genomad, genomad_output, genomad_db_path])
        else:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running geNomad (normal mode) on {sample_name}...\033[0m")
            subprocess.run(['genomad', 'end-to-end', '--cleanup', '--splits', '8', '-t', str(args["threads"]), input_for_genomad, genomad_output, genomad_db_path])

def run_checkv_first_run(sample_name, genomad_output, checkv_db_path, checkv_sample_directory, args):
    checkv_input_files = glob.glob(os.path.join(genomad_output, '**/*_virus.fna'))
    checkv_input = checkv_input_files[0]
    checkv_output = os.path.join(checkv_sample_directory, f'{sample_name}_Viruses_CheckV_Output')
    virus_checkv_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Viruses_CheckV_Output/quality_summary.tsv')
    
    if os.path.exists(virus_checkv_file) and not args["force_checkv"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping CheckV for {sample_name}, {checkv_output} folder already exists, and force_checkv not provided...\033[0m")
        df_virus_CHECKV = pd.read_csv(virus_checkv_file, sep='\t', usecols=['contig_id', 'contig_length', 'provirus', 'proviral_length', 'gene_count', 'viral_genes',
                                                                         'host_genes', 'checkv_quality', 'miuvig_quality', 'completeness', 'completeness_method', 'kmer_freq'])
    else:
        if getsize(checkv_input) == 0:
            print(f"\n\033[1m{step_counter.print_main_step()}: Skipping CheckV {checkv_input} because it is empty...\033[0m")
            # Create output folder for skipped files
            empty_checkv_output = os.path.join(checkv_output)
            os.makedirs(empty_checkv_output, exist_ok=True)
            # Write skipped_files.txt in the skipped folder
            with open(os.path.join(empty_checkv_output, 'Virus_CheckV_logfile.txt'), 'a') as f:
                f.write('Creating empty virus.fna and quality_summary.tsv')
            # Create empty viruses.fna and proviruses.fna files
            with open(os.path.join(empty_checkv_output, 'viruses.fna'), 'w') as f:
                pass
            with open(os.path.join(empty_checkv_output, 'proviruses.fna'), 'w') as f:
                pass
            with open(os.path.join(empty_checkv_output, 'quality_summary.tsv'), 'w') as f:
                f.write('contig_id\tcontig_length\tprovirus\tproviral_length\tgene_count\tviral_genes\thost_genes\tcheckv_quality\tmiuvig_quality\tcompleteness\tcompleteness_method\tcontamination\tkmer_freq\twarnings\n')
            df_virus_CHECKV = pd.read_csv(virus_checkv_file, sep='\t', usecols=['contig_id', 'contig_length', 'provirus', 'proviral_length', 'gene_count', 'viral_genes',
                                                                         'host_genes', 'checkv_quality', 'miuvig_quality', 'completeness', 'completeness_method', 'kmer_freq'])
        else:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running CheckV on {sample_name}...\033[0m")
            subprocess.run(['checkv', 'end_to_end', checkv_input, checkv_output, '-d', checkv_db_path, '-t', str(args["threads"]), '--remove_tmp'])
            genomad_proviruses_input = os.path.join(checkv_output, 'proviruses.fna')
            new_genomad_proviruses_input = genomad_proviruses_input + '.new'
            with open(genomad_proviruses_input, 'r') as f_in, open(new_genomad_proviruses_input, 'w') as f_out:
                for line in f_in:
                    if line.startswith('>'):
                        # Adjust contig name here, to reflect actual coordinates if this is a double provirus (Genomad provirus trimmed by CheckV)
                        new_header = rename_sequence_coord_checkv(line.strip().replace(' ', '|'))
                        f_out.write(new_header + '\n')
                    else:
                        f_out.write(line)
            # Replace the original file with the modified file
            os.replace(new_genomad_proviruses_input, genomad_proviruses_input)
            df_virus_CHECKV = pd.read_csv(virus_checkv_file, sep='\t', usecols=['contig_id', 'contig_length', 'provirus', 'proviral_length', 'gene_count', 'viral_genes',
                                                                         'host_genes', 'checkv_quality', 'miuvig_quality', 'completeness', 'completeness_method', 'kmer_freq'])

    return df_virus_CHECKV

def run_genomad_second_run(sample_name, checkv_output, genomad_sample_directory, args, genomad_db_path):
    genomad_proviruses_input = os.path.join(checkv_output, 'proviruses.fna')
    genomad_proviruses_output = os.path.join(genomad_sample_directory, f'{sample_name}_Proviruses_Genomad_Output')

    if os.path.exists(genomad_proviruses_output) and not args["force_genomad"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping Genomad on proviruses for {sample_name}, {genomad_proviruses_output} folder already exists, and force_genomad not provided...\033[0m")
    else:
        if getsize(genomad_proviruses_input) == 0:
            print(f"\n\033[1m{step_counter.print_main_step()}: Skipping Genomad on proviruses {genomad_proviruses_input} because it is empty...\033[0m")
            # Create output folder for skipped files
            empty_genomad_proviruses_output = os.path.join(genomad_proviruses_output, 'proviruses_summary')
            os.makedirs(empty_genomad_proviruses_output, exist_ok=True)
            # Write skipped_files.txt in the skipped folder
            with open(os.path.join(empty_genomad_proviruses_output, 'Provirus_Genomad_logfile.txt'), 'a') as f:
                f.write('Creating empty proviruses_virus.fna and proviruses_virus_summary.tsv')
            # Create empty viruses.fna file
            with open(os.path.join(empty_genomad_proviruses_output, 'proviruses_virus.fna'), 'w') as f:
                pass
            with open(os.path.join(empty_genomad_proviruses_output, 'proviruses_virus_summary.tsv'), 'w') as f:
                f.write('seq_name\tlength\ttopology\tcoordinates\tn_genes\tgenetic_code\tvirus_score\tfdr\tn_hallmarks\tmarker_enrichment\ttaxonomy\n')
        else:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running Genomad on proviruses {sample_name}...\033[0m")
            subprocess.run(['genomad', 'end-to-end', '--min-virus-hallmarks-short-seqs', '1', '--min-score', '0.5', '--cleanup', '--splits', '8', '-t', str(args["threads"]), genomad_proviruses_input, genomad_proviruses_output, genomad_db_path])

def run_checkv_second_run(sample_name, checkv_sample_directory, genomad_proviruses_output, args, checkv_db_path):
    checkv_proviruses_input = os.path.join(genomad_proviruses_output, 'proviruses_summary', 'proviruses_virus.fna')
    checkv_proviruses_output = os.path.join(checkv_sample_directory, f'{sample_name}_Proviruses_CheckV_Output')
    provirus_checkv_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Proviruses_CheckV_Output/quality_summary.tsv')

    if os.path.exists(provirus_checkv_file) and not args["force_checkv"]:
        print(f"\n\033[1m{step_counter.print_main_step()}: Skipping CheckV on proviruses for {sample_name}, {checkv_proviruses_output} folder already exists, and force_checkv not provided...\033[0m")
    else:
        if getsize(checkv_proviruses_input) == 0:
            print(f"\n\033[1m{step_counter.print_main_step()}: Skipping CheckV on {checkv_proviruses_input} because it is empty...\033[0m")
            # Create output folder for skipped files
            os.makedirs(checkv_proviruses_output, exist_ok=True)
            # Write skipped_files.txt in the skipped folder
            with open(os.path.join(checkv_proviruses_output, 'Provirus_CheckV_logfile.txt'), 'a') as f:
                f.write('Creating empty viruses.fna and quality_summary.tsv')
            # Create empty viruses.fna file
            with open(os.path.join(checkv_proviruses_output, 'viruses.fna'), 'w') as f:
                pass
            with open(os.path.join(checkv_proviruses_output, 'quality_summary.tsv'), 'w') as f:
                f.write('contig_id\tcontig_length\tprovirus\tproviral_length\tgene_count\tviral_genes\thost_genes\tcheckv_quality\tmiuvig_quality\tcompleteness\tcompleteness_method\tcontamination\tkmer_freq\twarnings\n')
        else:
            print(f"\n\033[1m{step_counter.print_main_step()}: Running CheckV on proviruses {sample_name}...\033[0m")
            subprocess.run(['checkv', 'end_to_end', checkv_proviruses_input, checkv_proviruses_output, '-d', checkv_db_path, '-t', str(args["threads"]), '--remove_tmp'])

        # Open the quality summary file for reading
        with open(provirus_checkv_file, 'r') as f:
            lines = f.readlines()

        # Find the index of the 'provirus' column
        header = lines[0].strip().split('\t')
        provirus_col_index = header.index('provirus')

        # Modify the lines where provirus is 'No' to 'Yes'
        for i in range(1, len(lines)):
            fields = lines[i].strip().split('\t')
            if fields[provirus_col_index] == 'No':
                fields[provirus_col_index] = 'Yes'
                lines[i] = '\t'.join(fields) + '\n'

        # Write the modified lines to the original file
        with open(provirus_checkv_file, 'w') as f:
            f.writelines(lines)

def create_tables(sample_name, df_virus_CHECKV, df_provirus_CHECKV, df_virus_GENOMAD, df_provirus_GENOMAD, args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating merged unfiltered virus-provirus Genomad-CheckV output tables for {sample_name}...\033[0m")
    
    # Construct the path to the ICTV_Taxonomy_List.tsv file
    script_directory = os.path.dirname(os.path.abspath(__file__))
    # Get the directory of the mvp directory
    mvp_directory = os.path.dirname(script_directory)
    # Load the ICTV Taxonomy List file
    ictv_taxonomy_file = os.path.join(mvp_directory, 'data', 'ICTV_Taxonomy_List.tsv')
    # Read the ICTV Taxonomy List file
    ictv_taxonomy_df = pd.read_csv(ictv_taxonomy_file, sep='\t')

    # Filter out provirus lines in df_virus_CHECKV
    df_virus_CHECKV = df_virus_CHECKV[df_virus_CHECKV['provirus'] != 'Yes']

    # Merge the filtered virus CheckV and proviruses CheckV DataFrames
    checkv_merged_df = pd.concat([df_virus_CHECKV, df_provirus_CHECKV], ignore_index=True)

    # Merge the virus Genomad and proviruses Genomad DataFrames
    genomad_merged_df = pd.concat([df_virus_GENOMAD, df_provirus_GENOMAD], ignore_index=True)

    # Merge the merged CheckV and the merged Genomad DataFrames
    checkv_genomad_merged_df = pd.merge(checkv_merged_df, genomad_merged_df, left_on='contig_id',right_on='seq_name', how='left')
    checkv_genomad_merged_df = checkv_genomad_merged_df.rename(columns={'contig_id': 'virus_id', 'contig_length': 'virus_length'}).drop('proviral_length', axis=1)

    # Cast the 'topology' column to string type
    checkv_genomad_merged_df['topology'] = checkv_genomad_merged_df['topology'].astype(str)
    checkv_genomad_merged_df['virus_id'] = checkv_genomad_merged_df['virus_id'].astype(str)

    # Copy over the number of genes from genomad ('n_genes) in the column 'gene_count' that is derived from CheckV
    checkv_genomad_merged_df['gene_count'] = checkv_genomad_merged_df['n_genes']

    # Remove the redundant column
    checkv_genomad_merged_df.drop(['n_genes'], inplace=True, axis=1)

    # Update the 'provirus' column for rows where 'topology' contains 'Provirus'
    checkv_genomad_merged_df.loc[checkv_genomad_merged_df['topology'].str.contains('Provirus'), 'provirus'] = 'Yes'

    # Change 'No terminal repeats' to 'Provirus' for rows where 'Yes' is present in the 'provirus' column
    checkv_genomad_merged_df.loc[checkv_genomad_merged_df['provirus'] == 'Yes', 'topology'] = checkv_genomad_merged_df.loc[checkv_genomad_merged_df['provirus'] == 'Yes', 'topology'].replace('No terminal repeats', 'Provirus')
    checkv_genomad_merged_df = checkv_genomad_merged_df.fillna('NA')

    # Apply functions to each row in checkv_genomad_merged_df
    checkv_genomad_merged_df['Genome type'] = checkv_genomad_merged_df['taxonomy'].apply(lambda x: taxon_to_genome_type(x, ictv_taxonomy_df))
    checkv_genomad_merged_df['Host type'] = checkv_genomad_merged_df['taxonomy'].apply(lambda x: taxon_to_host_type(x, ictv_taxonomy_df))

    # Exclude specific columns before saving the DataFrame
    columns_to_exclude = ['seq_name', 'topology']  
    checkv_genomad_merged_df = checkv_genomad_merged_df.drop(columns=columns_to_exclude)

    # Save the merged Checkv and Genomad file
    checkv_genomad_merged_output_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'MVP_01_{sample_name}_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')

    # Add a 'Sample' column with the current sample name
    checkv_genomad_merged_df.insert(0, 'Sample', sample_name)

    # Apply renaming logic to the 'virus_id' column
    checkv_genomad_merged_df['virus_id'] = checkv_genomad_merged_df['virus_id'].apply(rename_sequence)

    ## Note: the CheckV provirus would already be "clean" so this renaming should not affect them
    checkv_genomad_merged_df.to_csv(checkv_genomad_merged_output_file, sep='\t', index=False)

def generate_summary_report(args, sample_name, formatted_start_time, formatted_end_time, duration, genomad_sample_directory):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating summary reports for {sample_name}...\033[0m")
   
    # Define the lines you want to add
    module_01_header = """****************************************************************************
******************               MODULE 01                ******************
****************************************************************************
"""

    # Create a dictionary to hold argument descriptions and their default values
    argument_defaults = {
        '--input': args["input"],
        '--metadata': args["metadata"],
        '--genomad-relaxed': args["genomad_relaxed"],
        '--genomad-conservative': args["genomad_conservative"],
        '--sample-group': args["sample_group"],
        '--skip_modify_assemblies': args["skip_modify_assemblies"],
        '--min-seq-size': args["min_seq_size"],
        '--threads': args["threads"]}

    # Write a summary line with script arguments and their default values
    summary_line = "01_run_genomad_checkv.py"
    for arg, default in argument_defaults.items():
        if default is not None:
            summary_line += f" {arg} {default}"

    # Specify the path for the new MVP_01_Summary_Report.txt in '01_GENOMAD' directory
    summary_report_01_path = os.path.join(genomad_sample_directory, str(sample_name) + '_MVP_01_Summary_Report.txt')

    # Write the modified content to the new summary report in '01_GENOMAD'
    formatted_end_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    duration = datetime.datetime.now() - datetime.datetime.now()
    with open(summary_report_01_path, 'w') as summary_report:
        summary_report.write(module_01_header + summary_line)
        summary_report.write(f"\n")
        summary_report.write(f"\nStart Time: {formatted_start_time}\n")
        summary_report.write(f"End time: {formatted_end_time}\n")
        summary_report.write(f"Running Time: {duration.total_seconds():.2f} seconds\n")

def main(args):
    # Capture start time
    mvp_start_time = datetime.datetime.now()
    mvp_formatted_start_time = mvp_start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {mvp_formatted_start_time}") 

    # read metadata file and map FASTQ files
    metadata = pd.read_csv(args["metadata"], sep='\t')

    # Determine the geNomad database path
    genomad_db_path = os.path.join(args["input"], '00_DATABASES', 'genomad_db')
    if args["genomad_db_path"]:
        genomad_db_path = args["genomad_db_path"]

    # Determine the checkV database path
    checkv_db_path = os.path.join(args["input"], '00_DATABASES', 'checkv-db-v1.5')
    if args["checkv_db_path"]:
        checkv_db_path = args["checkv_db_path"]

    # Create a new directory for modified and filtered assembly files
    if not args["skip_modify_assemblies"]:
        modified_assembly_dir = os.path.join(args["input"], '00_MODIFIED_ASSEMBLY_FILES')
        os.makedirs(modified_assembly_dir, exist_ok=True)
    else: 
        modified_assembly_dir = None

    # Process rows based on specific sample numbers or all rows
    specific_sample_numbers = []
    if args["sample_group"]:
        specific_sample_numbers = [int(num) for num in args["sample_group"].split(',') if num.strip()]


    for row_index, (_, row) in enumerate(metadata.iterrows(), start=1):
        input_assembly_file = row['Assembly_Path']
        sample_number = row['Sample_number']
        sample_name = row['Sample']

        # Check if the current sample number is in the list of specific sample numbers
        if specific_sample_numbers and sample_number not in specific_sample_numbers:
            continue # Skip to the next iteration

        step_counter.reset()

        # Capture start time
        start_time = datetime.datetime.now()
        formatted_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")

        print(f"\nStart Time of processing {sample_name}: {formatted_start_time}") 

        input_for_genomad = modify_assembly_files(input_assembly_file, sample_name, modified_assembly_dir, args)

        # Create directories for the current sample in 01_GENOMAD and 02_CHECK_V
        genomad_sample_directory = os.path.join(args["input"], '01_GENOMAD', str(sample_name))
        os.makedirs(genomad_sample_directory, exist_ok=True)
        checkv_sample_directory = os.path.join(args["input"], '02_CHECK_V', str(sample_name))
        os.makedirs(checkv_sample_directory, exist_ok=True)

        genomad_output = os.path.join(genomad_sample_directory, f'{sample_name}_Viruses_Genomad_Output')
        checkv_output = os.path.join(checkv_sample_directory, f'{sample_name}_Viruses_CheckV_Output')
        genomad_proviruses_output = os.path.join(genomad_sample_directory, f'{sample_name}_Proviruses_Genomad_Output')

        run_genomad_first_run(sample_name, input_for_genomad, genomad_output, args, genomad_db_path)

        virus_genomad_file = glob.glob(os.path.join(args["input"], '01_GENOMAD', str(sample_name), f'{sample_name}_Viruses_Genomad_Output/*/*_virus_summary.tsv'))[0]
        df_virus_GENOMAD = pd.read_csv(virus_genomad_file, sep='\t', usecols=['seq_name', 'topology', 'n_genes', 'genetic_code', 'virus_score', 'n_hallmarks', 'marker_enrichment', 'taxonomy', 'coordinates'])
       
        run_checkv_first_run(sample_name, genomad_output, checkv_db_path, checkv_sample_directory, args)
        virus_checkv_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Viruses_CheckV_Output/quality_summary.tsv')
        df_virus_CHECKV = pd.read_csv(virus_checkv_file, sep='\t', usecols=['contig_id', 'contig_length', 'provirus', 'proviral_length', 'gene_count', 'viral_genes',
                                                                         'host_genes', 'checkv_quality', 'miuvig_quality', 'completeness', 'completeness_method', 'kmer_freq'])
      
        run_genomad_second_run(sample_name, checkv_output, genomad_sample_directory, args, genomad_db_path)
        provirus_genomad_file = glob.glob(os.path.join(args["input"], '01_GENOMAD', str(sample_name), f'{sample_name}_Proviruses_Genomad_Output/*/proviruses_virus_summary.tsv'))[0]
        df_provirus_GENOMAD = pd.read_csv(provirus_genomad_file, sep='\t', usecols=['seq_name', 'topology', 'n_genes', 'genetic_code', 'virus_score', 'n_hallmarks', 'marker_enrichment', 'taxonomy', 'coordinates'])
        # Correct the coordinates, since these are actually proviruses we reprocessed
        df_provirus_GENOMAD['coordinates'] = df_provirus_GENOMAD['seq_name'].apply(lambda x: correct_coordinates(x))

        run_checkv_second_run(sample_name, checkv_sample_directory, genomad_proviruses_output, args, checkv_db_path)
        provirus_checkv_file = os.path.join(args["input"], '02_CHECK_V', str(sample_name), f'{sample_name}_Proviruses_CheckV_Output/quality_summary.tsv')
        df_provirus_CHECKV = pd.read_csv(provirus_checkv_file, sep='\t', usecols=['contig_id', 'contig_length', 'provirus', 'proviral_length', 'gene_count', 'viral_genes', 'host_genes', 'checkv_quality', 'miuvig_quality', 'completeness', 'completeness_method', 'kmer_freq'])
     
        create_tables(sample_name, df_virus_CHECKV, df_provirus_CHECKV, df_virus_GENOMAD, df_provirus_GENOMAD, args)

        # Capture end time
        end_time = datetime.datetime.now()
        formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

        # Calculate duration
        duration = end_time - start_time

        generate_summary_report(args, sample_name, formatted_start_time, formatted_end_time, duration, genomad_sample_directory)

        print(f"\nEnd Time of processing {sample_name}: {formatted_end_time}")
        print(f"Running Time of processing {sample_name}: {duration.total_seconds():.2f} seconds")
        
    # Capture end time
    mvp_end_time = datetime.datetime.now()
    mvp_formatted_end_time = mvp_end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    mvp_duration = mvp_end_time - mvp_start_time
    
    message1 = "\033[1mModule 01 finished: geNomad and CheckV executed successfully!\033[0m\n"
    message2 = "geNomad and CheckV output directories for both viruses and proviruses have been generated."
    message3 = "You can look in <sample_name>_virus_summary.tsv, proviruses_virus_summary.tsv, and quality_summary.tsv tabular files for a summary of geNomad and CheckV outputs."
    message4 = "\n\033[1mYou can now proceed to the next step of the MVP script: Module 02!\033[0m"
    line_of_stars = '*' * len(message2)
    print()
    print(line_of_stars)
    print(message1)
    print(f"Start Time: {mvp_formatted_start_time}") 
    print(f"End Time: {mvp_formatted_end_time}")
    print(f"Running Time: {mvp_duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(message4)
    print(line_of_stars)
    print()

    print("Please don't forget to cite MVP and the following softwares used by this module:")
    print("- Camargo, A. P., Roux, S., Schulz, F., Babinski, M., Xu, Y., Hu, B., Chain, P. S. G., Nayfach, S., & Kyrpides, N. C. Identification of mobile genetic elements with geNomad. Nat Biotechnol, 1-10 (2023). https://doi.org/10.1038/s41587-023-01953-y")
    print("\n- Nayfach, S., Camargo, A.P., Schulz, F. et al. CheckV assesses the quality and completeness of metagenome-assembled viral genomes. Nat Biotechnol 39, 578–585 (2021). https://doi.org/10.1038/s41587-020-00774-7\n")



#!/usr/bin/env python

import time, resource, platform, sys, argparse, gzip

def parse_seqs(path):
	handle = gzip.open(path) if path.split('.')[-1] == 'gz' else open(path)
	id = next(handle).split()[0][1:]
	seq = ''
	for line in handle:
		if line[0] == '>':
			yield id, seq
			id = line.split()[0][1:]
			seq = ''
		else:
			seq += line.rstrip()
	yield id, seq
	handle.close()

def log_time(start):
	current_time = time.time()
	program_time = round(current_time - start, 2)
	peak_ram = round(max_mem_usage(), 2)
	print("time: %s seconds, peak RAM: %s GB" % (program_time, peak_ram))

def max_mem_usage():
	""" Return max mem usage (Gb) of self and child processes """
	max_mem_self = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
	max_mem_child = resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
	if platform.system() == 'Linux':
		return round((max_mem_self + max_mem_child)/float(1e6), 2)
	else:
		return round((max_mem_self + max_mem_child)/float(1e9), 2)

def parse_arguments():
	parser = argparse.ArgumentParser(
		formatter_class=argparse.RawTextHelpFormatter,
		usage=argparse.SUPPRESS,
		description="Centroid based sequence clustering"
	)
	parser.add_argument('--fna', type=str, required=True, metavar='PATH',
		help="""Path to nucleotide sequences""")
	parser.add_argument('--ani', type=str, required=True, metavar='PATH',
		help="""Path to tab-delimited file with fields: [qname, tname, num_alns, ani, qcov, tcov]""")
	parser.add_argument('--out', type=str, required=True, metavar='BASENAME',
		help="""Path to output file""")
	parser.add_argument('--exclude', type=str, metavar='PATH',
		help="""Path to list of sequence ids to exclude from clustering""")
	parser.add_argument('--keep', type=str, metavar='PATH',
		help="""Path to list of sequence ids to keep from clustering""")
	parser.add_argument('--min_ani', type=float, metavar='FLOAT', default=95,
		help="""Minimum average nucleotide identity (0...100, default=95)""")
	parser.add_argument('--min_qcov', type=float, metavar='FLOAT', default=0,
		help="""Minimum alignment coverage of longer sequence (0...100, default=0)""")
	parser.add_argument('--min_tcov', type=float, metavar='FLOAT', default=85,
		help="""Minimum alignment coverage of shorter sequence (0...100, default=85)""")
	parser.add_argument('--min_length', type=float, metavar='INT', default=1,
		help="""Minimum sequence length (default=1)""")
	return vars(parser.parse_args())

# main
start = time.time()

# args
args = parse_arguments()

# list seqs, sorted by length
print("\nreading sequences...")
seqs = {}
exclude = set([_.rstrip() for _ in open(args['exclude'])]) if args['exclude'] else None
keep = set([_.rstrip() for _ in open(args['keep'])]) if args['keep'] else None
for index, r in enumerate(parse_seqs(args['fna'])):
	id, seq  = r
	if len(seq) < args['min_length']:
		continue
	elif exclude and id in exclude:
		continue
	elif keep and id not in keep:
		continue
	else:
		seqs[id] = len(seq)
seqs = [x[0] for x in sorted(seqs.items(), key=lambda x: x[1], reverse=True)]
print("%s sequences retained from fna" % len(seqs))
log_time(start)

# store edges
print("\nstoring edges...")
num_edges = 0
edges = dict([(x,[]) for x in seqs])
handle = gzip.open(args['ani']) if args['ani'].split('.')[-1] == 'gz' else open(args['ani'])
for index, line in enumerate(handle):
	qname, tname, num_alns, ani, qcov, tcov = line.split()
	if qname == tname:
		continue
	elif qname not in edges or tname not in edges:
		continue
	elif float(qcov) < args['min_qcov'] or float(tcov) < args['min_tcov'] or float(ani) < args['min_ani']:
		continue
	edges[qname].append(tname)
	num_edges += 1
#	if not num_edges % 1000000:
#		current_time = time.time()
#		program_time = round(current_time - start, 2)
#		peak_ram = round(max_mem_usage(), 2)
#		print("num edges: %s, seconds: %s, GB: %s" % (num_edges, program_time, peak_ram))
handle.close()
print("%s edges retained from blastani" % num_edges)
print("%s edges currently stored" % sum([len(_) for _ in edges.values()]))
log_time(start)

# cluster
print("\nclustering...")
clust_to_seqs = {}
seq_to_clust = {}
# loop over seqs in sorted order
for seq_id in seqs:
	# seq already been assigned; cant be centroid
	if seq_id in seq_to_clust:
		continue
	# seq is centroid for new cluster
	else:
		# add self to cluster
		clust_to_seqs[seq_id] = [seq_id]
		seq_to_clust[seq_id] = seq_id
		# update with cluster members
		for mem_id in edges[seq_id]:
			if mem_id not in seq_to_clust:
				clust_to_seqs[seq_id].append(mem_id)
				seq_to_clust[mem_id] = seq_id
print("%s total clusters" % len(clust_to_seqs))
log_time(start)

# write
print("\nwriting clusters...")
out = open(args['out'], 'w')
for seq_id, mem_ids in clust_to_seqs.items():
	out.write(seq_id + '\t' + ','.join(mem_ids)+'\n')
log_time(start)

